#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides a set of basic taurus widgets based on QLineEdit
"""

from builtins import bytes
from builtins import str
import sys
import numpy
from taurus.external.qt import Qt
from taurus.core.units import Quantity
from taurus.qt.qtgui.base import TaurusBaseWritableWidget
from taurus.qt.qtgui.util import PintValidator
from taurus.core import DataType, DataFormat, TaurusEventType

__all__ = ["TaurusValueLineEdit"]

__docformat__ = 'restructuredtext'


class TaurusValueLineEdit(Qt.QLineEdit, TaurusBaseWritableWidget):
    """
    A taurus-aware ``QLineEdit``. It will display the value (or fragment
    of the value) referenced by its model. It is a
    :class:`~taurus.qt.qtgui.base.TaurusBaseWritableWidget`
    and as such it does not apply the changes straight away to the model
    (unless ``autoApply`` is enabled), but instead shows that there are pending
    operations which can be applied by pressing "ENTER".

    When used with numerical value Attributes as its model, it provides some
    extended behaviour:

    - It represents out-of-limit values using different colours (for warning,
      range, invalid,...)
    - It uses a validator that is range-aware
    - The mouse wheel and keyboard arrows can be enabled for doing value
      increments

    .. note::
        when used with models whose value is a pint `Quantity`, the text
        is parsed by pint and therefore one can write e.g. `2 3 mm` which is
        equivalent to `6 mm` !

    """

    _bytesEncoding = sys.stdin.encoding

    def __init__(self, qt_parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QLineEdit, qt_parent)
        self.call__init__(TaurusBaseWritableWidget,
                          name, designMode=designMode)
        self._enableWheelEvent = False
        self._last_value = None
        self._singleStep = 1.

        self.setAlignment(Qt.Qt.AlignRight)
        self.setValidator(None)

        self.textChanged.connect(self.notifyValueChanged)
        self.returnPressed.connect(self.writeValue)
        self.valueChangedSignal.connect(self.updatePendingOperations)
        self.editingFinished.connect(self._onEditingFinished)

    def _updateValidator(self, value):
        """This method sets a validator depending on the data type"""
        val = None
        if isinstance(value.wvalue, Quantity):
            val = self.validator()
            if not isinstance(val, PintValidator):
                val = PintValidator(self)
                self.setValidator(val)
            attr = self.getModelObj()
            bottom, top = attr.range
            if bottom != val.bottom:
                val.setBottom(bottom)
            if top != val.top:
                val.setTop(top)
            units = value.wvalue.units
            if units != val.units:
                val.setUnits(units)

        # @TODO Other validators can be configured for other types
        #       (e.g. with string lengths, tango names,...)
        else:
            self.setValidator(None)
            self.debug("Validator disabled")
        return val

    def __decimalDigits(self, fmt):
        """returns the number of decimal digits from a format string
        (or None if they are not defined)"""
        try:
            if fmt[-1].lower() in ['f', 'g'] and '.' in fmt:
                return int(fmt[:-1].split('.')[-1])
            else:
                return None
        except:
            return None

    def _onEditingFinished(self):
        """slot for performing autoapply only when edition is finished"""
        if self._autoApply:
            self.writeValue()

    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
    # TaurusBaseWritableWidget overwriting
    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
    def notifyValueChanged(self, *args):
        """reimplement to avoid autoapply on every partial edition"""
        self.emitValueChanged()

    def handleEvent(self, evt_src, evt_type, evt_value):

        # handle the case in which the line edit is not yet initialized
        if self._last_value is None:
            try:
                value = self.getModelObj().read(cache=True)
                self._updateValidator(value)
                self.setValue(value.wvalue)
            except Exception as e:
                self.info('Failed attempt to initialize value: %r', e)

        self.setEnabled(evt_type != TaurusEventType.Error)

        if evt_type in (TaurusEventType.Change, TaurusEventType.Periodic):
            self._updateValidator(evt_value)

        TaurusBaseWritableWidget.handleEvent(
            self, evt_src, evt_type, evt_value)

        if evt_type == TaurusEventType.Error:
            self.updateStyle()

    def isTextValid(self):
        """
        Validates current text

        :return: (bool) Returns False if there is a validator and the current
                 text is not Acceptable. Returns True otherwise.
        """
        val = self.validator()
        if val is None:
            return True
        return val.validate(str(self.text()), 0)[0] == val.Acceptable

    def updateStyle(self):
        """Reimplemented from :class:`TaurusBaseWritableWidget`"""
        TaurusBaseWritableWidget.updateStyle(self)

        value = self.getValue()
        
        if value is None or not self.isTextValid() or not self.isEnabled():
            # invalid value
            color, weight = 'gray', 'normal'
        else:
            # check if there are pending operations
            if self.hasPendingOperations():
                color, weight = 'blue', 'bold'
            else:
                color, weight = 'black', 'normal'
            # also check alarms (if applicable)
            modelObj = self.getModelObj()
            if modelObj and modelObj.type in [DataType.Integer,
                                              DataType.Float]:
                min_, max_ = modelObj.alarms
                if ((min_ is not None and value < min_) or
                        (max_ is not None and value > max_)):
                    color = 'orange'
        # apply style
        style = ('TaurusValueLineEdit {color: %s; font-weight: %s}' %
                 (color, weight))
        self.setStyleSheet(style)

    def wheelEvent(self, evt):
        """Wheel event handler"""
        if not self.getEnableWheelEvent() or Qt.QLineEdit.isReadOnly(self):
            return Qt.QLineEdit.wheelEvent(self, evt)
        model = self.getModelObj()
        if model is None or not model.isNumeric():
            return Qt.QLineEdit.wheelEvent(self, evt)

        evt.accept()
        numDegrees = evt.delta() // 8
        numSteps = numDegrees // 15
        self._stepBy(numSteps)

    def keyPressEvent(self, evt):
        """Key press event handler"""
        if evt.key() in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            Qt.QLineEdit.keyPressEvent(self, evt)
            evt.accept()
            return
        if Qt.QLineEdit.isReadOnly(self):
            return Qt.QLineEdit.keyPressEvent(self, evt)
        model = self.getModelObj()
        if model is None or not model.isNumeric():
            return Qt.QLineEdit.keyPressEvent(self, evt)

        if evt.key() == Qt.Qt.Key_Up:
            numSteps = 1
        elif evt.key() == Qt.Qt.Key_Down:
            numSteps = -1
        else:
            return Qt.QLineEdit.keyPressEvent(self, evt)

        evt.accept()
        self._stepBy(numSteps)

    def _stepBy(self, steps):
        value = self.getValue()
        self.setValue(value + Quantity(steps * self._singleStep, value.units))

        if self.getAutoApply():
            self.editingFinished.emit()
        else:
            kmods = Qt.QCoreApplication.instance().keyboardModifiers()
            controlpressed = bool(kmods & Qt.Qt.ControlModifier)
            if controlpressed:
                self.writeValue(forceApply=True)

    def setValue(self, v):
        """Set the displayed text from a given value object"""
        # Support displaying the value without units (enabled by fragment)
        # Other fragments are ignored by setValue
        if self.modelFragmentName == "wvalue.magnitude":
            try:
                validator = self.validator()
                if validator is None:
                    value = self.getModelValueObj()
                    validator = self._updateValidator(value)
                units = validator.units
                v = v.to(units).magnitude
            except Exception as e:
                self.debug('Cannot enforce fragment. Reason: %r', e)
        self._last_value = v
        self.setText(str(self.displayValue(v)).strip())

    def getValue(self):
        text = self.text()
        model_obj = self.getModelObj()
        if model_obj is None:
            return None
        val = self.validator()
        try:
            model_type = model_obj.type
            model_format = model_obj.data_format
            if model_type in [DataType.Integer, DataType.Float]:
                try:
                    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    # workaround for https://github.com/hgrecco/pint/issues/614
                    text = text.lstrip('0') or '0'
                    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    q = Quantity(text)
                    # allow implicit units (assume wvalue.units implicitly)
                    if q.unitless:
                        q = Quantity(q.magnitude, val.units)
                    return q
                except:
                    return None
            elif model_type == DataType.Boolean:
                if model_format == DataFormat._0D:
                    return bool(int(eval(text)))
                else:
                    return numpy.array(eval(text), dtype=int).astype(bool)
            elif model_type == DataType.String:
                if model_format == DataFormat._0D:
                    return str(text)
                else:
                    return numpy.array(eval(text), dtype=str).tolist()
            elif model_type == DataType.Bytes:
                return bytes(text, self._bytesEncoding)
            else:
                raise TypeError('Unsupported model type "%s"' % model_type)
        except Exception as e:
            msg = 'Cannot return value for "%s". Reason: %r'
            if text in (str(None), self.getNoneValue()):
                self.debug(msg, text, e)
            else:
                self.warning(msg, text, e)
            return None

    def setEnableWheelEvent(self, b):
        self._enableWheelEvent = b

    def getEnableWheelEvent(self):
        return self._enableWheelEvent

    def resetEnableWheelEvent(self):
        self.setEnableWheelEvent(False)

    def getSingleStep(self):
        return self._singleStep

    def setSingleStep(self, step):
        self._singleStep = step

    def resetSingleStep(self):
        self.setSingleStep(1.0)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWritableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.input'
        ret['icon'] = "designer:lineedit.png"
        return ret

    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
    # QT properties
    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

    model = Qt.pyqtProperty("QString", TaurusBaseWritableWidget.getModel,
                            TaurusBaseWritableWidget.setModel,
                            TaurusBaseWritableWidget.resetModel)

    useParentModel = Qt.pyqtProperty(
        "bool", TaurusBaseWritableWidget.getUseParentModel,
        TaurusBaseWritableWidget.setUseParentModel,
        TaurusBaseWritableWidget.resetUseParentModel)

    autoApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getAutoApply,
                                TaurusBaseWritableWidget.setAutoApply,
                                TaurusBaseWritableWidget.resetAutoApply)

    forcedApply = Qt.pyqtProperty("bool",
                                  TaurusBaseWritableWidget.getForcedApply,
                                  TaurusBaseWritableWidget.setForcedApply,
                                  TaurusBaseWritableWidget.resetForcedApply)

    enableWheelEvent = Qt.pyqtProperty("bool", getEnableWheelEvent,
                                       setEnableWheelEvent,
                                       resetEnableWheelEvent)


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <full_attribute_name(s)>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus lineedit demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()

    form = Qt.QWidget()
    layout = Qt.QVBoxLayout()
    form.setLayout(layout)
    if len(args) == 0:
        models = ['sys/tg_test/1/double_scalar', 'sys/tg_test/1/double_scalar']
    else:
        models = args
    for model in models:
        w = TaurusValueLineEdit()
        w.setModel(model)
        layout.addWidget(w)
    form.resize(300, 50)
    form.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return form


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core.resource.test.test_resvalidator..."""

import os.path as osp

import taurus
from taurus.core.taurusexception import TaurusException
import unittest
from taurus.core.test import (valid, invalid, names,
                              AbstractNameValidatorTestCase)
from taurus.core.resource.resvalidator import (ResourceAuthorityNameValidator,
                                               ResourceDeviceNameValidator,
                                               ResourceAttributeNameValidator)


class _AbstractResNameValidatorTestCase(AbstractNameValidatorTestCase):
    """Abstract class for creating res validator test cases. Derived classes
    need to provide the `res_map` dictionary class member"""
    res_map = {}

    def setUp(self):
        unittest.TestCase.setUp(self)
        f = taurus.Factory('res')
        f.clear()  # make sure we use a clean res factory
        f.loadResource(self.res_map)

    def tearDown(self):
        f = taurus.Factory('res')
        f.clear()


@valid(name='res:foo02')
@valid(name='res:localhost')
@invalid(name='res:foo:10000')  # Invalid key
@invalid(name='res:10000')  # Invalid key
@invalid(name='res:127.0.0.1')  # Invalid key
@invalid(name='res:badtango_1', exceptionType=TaurusException)  # invalid value
class ResourceAuthorityValidatorTestCase(_AbstractResNameValidatorTestCase,
                                         unittest.TestCase):
    """
    Test for ResourceAuthorityNameValidator loading the resources
    from a dictionary.
    """
    validator = ResourceAuthorityNameValidator
    res_map = {'badtango_1': 'foo:10000',  # bad auth. Should start by "//"
               'foo02': 'tango://foo:10000',
               'localhost': 'eval://localhost'
               }


@valid(name='res:MyDev')
@valid(name='res:tangoDev1')
@valid(name='res:tangoDev_bck')
@invalid(name='res:123')  # Invalid key
@invalid(name='res:wrong_dev')  # Invalid eval URI
class ResourceDeviceValidatorTestCase(_AbstractResNameValidatorTestCase,
                                      unittest.TestCase):
    """
    Test for ResourceDeviceNameValidator loading the resources
    from a dictionary.
    """
    validator = ResourceDeviceNameValidator
    res_map = {'MyDev': 'eval:@foo',
               'tangoDev1': 'tango://foo:10000/a/b/c',
               'tangoDev_bck': 'tango:a/b/c',
               'wrong_dev': 'eval://mydev',
               }


@names(name='MyAttr', out=('eval://localhost/@DefaultEvaluator/1', '1', '1'))
@names(name='foo', out=('eval://localhost/@Foo/True', '@Foo/True', 'True'))
@valid(name='res:MyAttr')
@valid(name='res:My_Attr')
@valid(name='res:attr_1')
@valid(name='res:attr_2')
@invalid(name='res:attr_tango_bck_1')
@valid(name='res:attr_tango_bck_1', strict=False)
@valid(name='res:attr_tango_bck_2')
@valid(name='res:attr_state')
@valid(name='res:attr1')
@valid(name='res:foo')
@valid(name='res:Foo')
@valid(name='res:res_attr')
@invalid(name='res:1')
@invalid(name='res:1foo')
@invalid(name='res: foo')
@invalid(name='res:dev1')  # Is a device!
@invalid(name='res:dev2')  # Is a device!
@invalid(name='res:NotExist')  # Not existing reference!
class ResourceAttributeValidatorTestCase(_AbstractResNameValidatorTestCase,
                                         unittest.TestCase):
    """
    Test for ResourceAttributeNameValidator loading the resources
    from a dictionary.
    """
    validator = ResourceAttributeNameValidator
    res_map = {'MyAttr': 'eval:1',
               'My_Attr': 'eval:foo=1;bar=2;foo+bar',
               'attr_1': 'tango:a/b/c/d',
               'attr_2': 'a/b/c/d',
               'attr_tango_bck_1': 'tango://a/b/c/d',
               'attr_tango_bck_2': 'tango://foo:10000/a/b/c/d',
               'attr_state': 'a/b/c/state',
               'attr1': 'eval:"Hello_World!!"',
               'foo': 'eval:/@Foo/True',
               '1foo': 'eval:2',
               'Foo': 'eval:False',
               'res_attr': 'res:attr1',
               'dev1': 'tango:a/b/c',  # invalid
               'dev2': 'eval:@foo',   # invalid
               }



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
RawDataChooser.py:  widget for importing RawData (from file or from a function)
"""

__all__ = ["QRawDataWidget"]

import numpy

from taurus.external.qt import Qt
from taurus.core.util.safeeval import SafeEvaluator
from taurus.qt.qtgui.util.ui import UILoadable


@UILoadable
class QRawDataWidget(Qt.QWidget):

    ReadFromFiles = Qt.pyqtSignal(int, int)
    AddCurve = Qt.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(QRawDataWidget, self).__init__(parent)
        self.loadUi()

        # connecttions
        self.openFilesBT.clicked.connect(self.onOpenFilesButtonClicked)
        self.addCurveBT.clicked.connect(self.onAddCurveButtonClicked)

        # set validators in LE's
        self.xFromLE.setValidator(Qt.QDoubleValidator(self))
        self.xToLE.setValidator(Qt.QDoubleValidator(self))
        self.xStepLE.setValidator(Qt.QDoubleValidator(self))

    def onOpenFilesButtonClicked(self):
        """ Emit a ReadFromFiles signal with the selected xcol and skiprows as parameters"""
        xcol = self.xcolSB.value()
        if xcol == self.xcolSB.minimum():
            xcol = None
        skiprows = self.headerSB.value()
        self.ReadFromFiles.emit(xcol, skiprows)

    def onAddCurveButtonClicked(self):
        """ Emit a AddCurve signal with a rawdata dictionary as a parameter.
        The rawdata dictionary is prepared from the from the GUI's selection."""
        rawdata = {}
        if self.xRangeRB.isChecked():
            rawdata['x'] = numpy.arange(float(self.xFromLE.text()), float(
                self.xToLE.text()), float(self.xStepLE.text()))
        else:
            sev = SafeEvaluator()
            try:
                rawdata['x'] = sev.eval(str(self.xValuesLE.text()))
            except:
                Qt.QMessageBox.warning(
                    self, 'Invalid x values' 'Cannot interpret the x values.\n Use Python expressions like "[1, 3 , 67]" or "arange(100)")')
                return
        rawdata['f(x)'] = str(self.f_xLE.text())
        self.AddCurve.emit(rawdata)

if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)
    form = QRawDataWidget()
    form.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################


"""
qdoublelist.py: Provides a generic dialog containing two list which can move
items from one to the other
"""

from __future__ import print_function

from builtins import str
from builtins import range

from taurus.external.qt import Qt
from taurus.qt.qtgui.util.ui import UILoadable


__all__ = ["QDoubleListDlg"]

__docformat__ = 'restructuredtext'


@UILoadable(with_ui='ui')
class QDoubleListDlg(Qt.QDialog):
    '''Generic dialog providing two lists. Items can be moved from one to the other
    '''
    #@todo: drag&drop is disabled because Qt<4.6 does not have QList.setDefaultDragAndDropMode()

    def __init__(self, parent=None, designMode=False, winTitle='', mainLabel='',
                 label1='', label2='', list1=None, list2=None):

        if list1 is None:
            list1 = []
        if list2 is None:
            list2 = []

        super(QDoubleListDlg, self).__init__(parent)
        self.loadUi()

        if winTitle:
            self.setWindowTitle(winTitle)
        self.ui.mainLabel.setText(mainLabel)
        self.ui.group1.setTitle(label1)
        self.ui.group2.setTitle(label2)

        self.setList1(list1)
        self.setList2(list2)

        self.ui.to1BT.clicked.connect(self.onTo1)
        self.ui.to2BT.clicked.connect(self.onTo2)

    def _moveItem(self, fromlist, tolist):
        selected = fromlist.selectedItems()
        for item in selected:
            fromlist.takeItem(fromlist.row(item))
            tolist.addItem(item)

    def setList1(self, list1):
        '''sets the items to be present in the first list

        :param list2: (seq<str>) a sequence of strings
        '''
        self.ui.list1.clear()
        self.ui.list1.addItems(list1)

    def setList2(self, list2):
        '''sets the items to be present in the second list

        :param list2: (seq<str>) a sequence of strings
        '''
        self.ui.list2.clear()
        self.ui.list2.addItems(list2)

    def onTo1(self, *args):
        '''slot to be called when the "To1" button is pressed'''
        self._moveItem(self.ui.list2, self.ui.list1)

    def onTo2(self, *args):
        '''slot to be called when the "To2" button is pressed'''
        self._moveItem(self.ui.list1, self.ui.list2)

    def getAll1(self):
        '''returns a copy the items in the first list

        :return: (list<str>)
        '''
        return [str(self.ui.list1.item(row).text()) for row in range(self.ui.list1.count())]

    def getAll2(self):
        '''returns a copy the items in the second list

        :return: (list<str>)
        '''
        return [str(self.ui.list2.item(row).text()) for row in range(self.ui.list2.count())]

    # note, for the moment we do not make it available in designer because it does not
    # behave well as a widget (only as a dialog) (e.g., it closes if ESC is pressed
#    @classmethod
#    def getQtDesignerPluginInfo(cls):
#        return {
#            'module' : 'taurus.qt.qtgui.panel',
#            'group' : 'Taurus Input',
#            #'icon' : "designer:ledred.png",
#            'container' : False,
#        }
#------------------------------------------------------------------------------


def main():
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)

    dlg = QDoubleListDlg(winTitle='foo', mainLabel='bla, bla', label1='1', label2='2',
                         list1=['11', '22'], list2=['123', '33'])
    result = dlg.exec_()

    print("Result", result)
    print("list1", dlg.getAll1())
    print("list2", dlg.getAll2())


if __name__ == "__main__":
    import sys
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
paneldescriptionwizard.py:
"""

from __future__ import print_function

from taurus.external.qt import Qt
import sys
import weakref
from taurus.qt.qtgui.taurusgui.utils import PanelDescription
from taurus.qt.qtgui.icon import getCachedPixmap
from taurus.qt.qtgui.input import GraphicalChoiceWidget
from taurus.qt.qtgui.panel import TaurusModelChooser
from taurus.qt.qtgui.base import TaurusBaseComponent, TaurusBaseWidget
from taurus.qt.qtcore.communication import SharedDataManager
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_LIST_MIME_TYPE
from taurus.qt.qtgui.util import TaurusWidgetFactory
from taurus.core.util.log import Logger
import inspect
import copy


__all__ = ["PanelDescriptionWizard"]


class ExpertWidgetChooserDlg(Qt.QDialog):
    CHOOSE_TYPE_TXT = '(choose type)'

    memberSelected = Qt.pyqtSignal(dict)

    def __init__(self, parent=None):
        Qt.QDialog.__init__(self, parent)

        self.setWindowTitle('Advanced panel type selection')

        layout1 = Qt.QHBoxLayout()
        layout2 = Qt.QHBoxLayout()
        layout = Qt.QVBoxLayout()

        # subwidgets
        self.moduleNameLE = Qt.QLineEdit()
        self.moduleNameLE.setValidator(Qt.QRegExpValidator(
            Qt.QRegExp('[a-zA-Z0-9\.\_]*'), self.moduleNameLE))
        self.membersCB = Qt.QComboBox()
        self.dlgBox = Qt.QDialogButtonBox(
            Qt.QDialogButtonBox.Ok | Qt.QDialogButtonBox.Cancel)
        self.dlgBox.button(Qt.QDialogButtonBox.Ok).setEnabled(False)

        # layout
        layout.addWidget(
            Qt.QLabel('Select the module and widget to use in the panel:'))
        layout1.addWidget(Qt.QLabel('Module'))
        layout1.addWidget(self.moduleNameLE)
        layout2.addWidget(Qt.QLabel('Class (or widget)'))
        layout2.addWidget(self.membersCB)
        layout.addLayout(layout1)
        layout.addLayout(layout2)
        layout.addWidget(self.dlgBox)
        self.setLayout(layout)

        # connections
        self.moduleNameLE.editingFinished.connect(self.onModuleSelected)
        self.moduleNameLE.textEdited.connect(self.onModuleEdited)
        self.membersCB.activated.connect(self.onMemberSelected)
        self.dlgBox.accepted.connect(self.accept)
        self.dlgBox.rejected.connect(self.reject)

    def onModuleEdited(self):
        self.dlgBox.button(Qt.QDialogButtonBox.Ok).setEnabled(False)
        self.module = None
        self.moduleNameLE.setStyleSheet('')
        self.membersCB.clear()

    def onModuleSelected(self):
        modulename = str(self.moduleNameLE.text())
        try:
            __import__(modulename)
            # We use this because __import__('x.y') returns x instead of y !!
            self.module = sys.modules[modulename]
            self.moduleNameLE.setStyleSheet('QLineEdit {color: green}')
        except Exception as e:
            Logger().debug(repr(e))
            self.moduleNameLE.setStyleSheet('QLineEdit {color: red}')
            return
        # inspect the module to find the members we want (classes or widgets
        # inheriting from QWidget)
        members = inspect.getmembers(self.module)
        classnames = sorted(
            [n for n, m in members if inspect.isclass(m) and issubclass(m, Qt.QWidget)])
        widgetnames = sorted(
            [n for n, m in members if isinstance(m, Qt.QWidget)])
        self.membersCB.clear()
        self.membersCB.addItem(self.CHOOSE_TYPE_TXT)
        self.membersCB.addItems(classnames)
        if classnames and widgetnames:
            self.membersCB.InsertSeparator(self.membersCB.count())
        self.membersCB.addItems(classnames)

    def onMemberSelected(self, text):
        if str(text) == self.CHOOSE_TYPE_TXT:
            return
        self.dlgBox.button(Qt.QDialogButtonBox.Ok).setEnabled(True)
        # emit a signal with a dictionary that can be used to initialize
        self.memberSelected.emit(self.getMemberDescription())

    def getMemberDescription(self):
        try:
            membername = str(self.membersCB.currentText())
            member = getattr(self.module, membername, None)
            result = {'modulename': self.module.__name__}
        except Exception as e:
            Logger().debug('Cannot get member description: %s', repr(e))
            return None
        if inspect.isclass(member):
            result['classname'] = membername
        else:
            result['widgetname'] = membername
        return result

    @staticmethod
    def getDialog():
        dlg = ExpertWidgetChooserDlg()
        dlg.exec_()
        return dlg.getMemberDescription(), (dlg.result() == dlg.Accepted)


class BlackListValidator(Qt.QValidator):

    stateChanged = Qt.pyqtSignal(int, int)

    def __init__(self, blackList=None, parent=None):
        Qt.QValidator.__init__(self, parent)
        if blackList is None:
            blackList = []
        self.blackList = blackList
        self._previousState = None
        # check the signature of the validate method (it changed from old to new versions)
        # see
        # :http://www.riverbankcomputing.co.uk/static/Docs/PyQt4/html/python_v3.html#qvalidator
        dummyValidator = Qt.QDoubleValidator(None)
        self._oldMode = len(dummyValidator.validate('', 0)) < 3

    def validate(self, input, pos):
        if str(input) in self.blackList:
            state = self.Intermediate
        else:
            state = self.Acceptable
        if state != self._previousState:
            self.stateChanged.emit(state, self._previousState)
            self._previousState = state
        if self._oldMode:  # for backwards compatibility with older versions of PyQt
            return state, pos
        else:
            return state, input, pos

# class NamePage(Qt.QWizardPage):
#
#    def __init__(self, parent = None):
#        Qt.QWizardPage.__init__(self, parent)
#
#        self.setTitle('Name')
#        self.setSubTitle('Choose a name for the new panel (it must not exist already)')
#
#        #contents
#        self.nameLE = Qt.QLineEdit()
#        self.registerField("panelname*",self.nameLE)
#        self.diagnosticLabel = Qt.QLabel('')
#
#        layout = Qt.QHBoxLayout()
#        layout.addWidget(Qt.QLabel("Panel Name"))
#        layout.addWidget(self.nameLE)
#        layout.addWidget(self.diagnosticLabel)
#        self.setLayout(layout)
#
#    def initializePage(self):
#        gui = self.wizard().getGui()
#        if hasattr(gui, 'getPanelNames'):
#            pnames = gui.getPanelNames()
#            v = BlackListValidator(blackList=pnames, parent=self.nameLE)
#            self.nameLE.setValidator(v)
#            self.connect(v, Qt.SIGNAL('stateChanged'),self._onValidatorStateChanged)
#
#    def _onValidatorStateChanged(self, state, previous):
#        if state == Qt.QValidator.Acceptable:
#            self.diagnosticLabel.setText('')
#        else:
#            self.diagnosticLabel.setText('<b>(Name already exists)</b>')
#


class WidgetPage(Qt.QWizardPage, TaurusBaseWidget):
    OTHER_TXT = 'Other...'
    defaultCandidates = ['TaurusForm', 'TaurusTrend', 'TaurusPlot',
                         'TaurusImageDialog', 'TaurusTrend2DDialog', 'TaurusNeXusBrowser',
                         'TaurusDbTreeWidget', 'TaurusArrayEditor',
                         'SardanaEditor', 'TaurusJDrawSynopticsView',
                         'TaurusDevicePanel']

    def __init__(self, parent=None, designMode=False, extraWidgets=None):
        Qt.QWizardPage.__init__(self, parent)
        TaurusBaseWidget.__init__(self, 'WidgetPage')
        if extraWidgets:
            customWidgets, customWidgetScreenshots = list(zip(*extraWidgets))
            pixmaps = {}
            for k, s in extraWidgets:
                if s is None:
                    pixmaps[k] = None
                else:
                    try:
                        pixmaps[k] = getCachedPixmap(s)
                        if pixmaps[k].isNull():
                            raise Exception('Invalid Pixmap')
                    except:
                        self.warning('Could not create pixmap from %s' % s)
                        pixmaps[k] = None

        else:
            customWidgets, customWidgetScreenshots = [], []
            pixmaps = {}
        self.setFinalPage(True)
        self.setTitle('Panel type')
        self.setSubTitle('Choose a name and type for the new panel')
        self.setButtonText(Qt.QWizard.NextButton, 'Advanced settings...')

        self.widgetDescription = {'widgetname': None,
                                  'modulename': None, 'classname': None}

        # name part
        self.nameLE = Qt.QLineEdit()
        self.registerField("panelname*", self.nameLE)
        self.diagnosticLabel = Qt.QLabel('')
        nameLayout = Qt.QHBoxLayout()
        nameLayout.addWidget(Qt.QLabel("Panel Name"))
        nameLayout.addWidget(self.nameLE)
        nameLayout.addWidget(self.diagnosticLabel)

        # contents
        available = TaurusWidgetFactory().getWidgetClassNames()
        choices = []
        row = []
        for cname in self.defaultCandidates + list(customWidgets):
            if cname in available or '.' in cname:
                row.append(cname)
                if cname not in pixmaps:
                    pixmaps[cname] = getCachedPixmap('snapshot:%s.png' % cname)
                if len(row) == 3:
                    choices.append(row)
                    row = []
        row.append(self.OTHER_TXT)
        choices.append(row)

        # defaultPixmap=getPixmap('logos:taurus.png')
        self.choiceWidget = GraphicalChoiceWidget(
            choices=choices, pixmaps=pixmaps)

        self.widgetTypeLB = Qt.QLabel("<b>Widget Type:</b>")

        self.choiceWidget.choiceMade.connect(self.onChoiceMade)

        layout = Qt.QVBoxLayout()
        layout.addLayout(nameLayout)
        layout.addWidget(self.choiceWidget)
        layout.addWidget(self.widgetTypeLB)
        self.setLayout(layout)

    def initializePage(self):
        gui = self.wizard().getGui()
        if hasattr(gui, 'getPanelNames'):
            pnames = gui.getPanelNames()
            v = BlackListValidator(blackList=pnames, parent=self.nameLE)
            self.nameLE.setValidator(v)
            v.stateChanged.connect(self._onValidatorStateChanged)

    def validatePage(self):
        paneldesc = self.wizard().getPanelDescription()
        if paneldesc is None:
            Qt.QMessageBox.information(self, 'You must choose a panel type',
                                       'Choose a panel type by clicking on one of the proposed types')
            return False
        try:
            w = paneldesc.getWidget()
            if not isinstance(w, Qt.QWidget):
                raise ValueError
            # set the name now because it might have changed since the
            # PanelDescription was created
            paneldesc.name = self.field('panelname')
            # allow the wizard to proceed
            return True
        except Exception as e:
            Qt.QMessageBox.warning(
                self, 'Invalid panel', 'The requested panel cannot be created. \nReason:\n%s' % repr(e))
            return False

    def _onValidatorStateChanged(self, state, previous):
        if state == Qt.QValidator.Acceptable:
            self.diagnosticLabel.setText('')
        else:
            self.diagnosticLabel.setText('<b>(Name already exists)</b>')

    def onChoiceMade(self, choice):
        if choice == self.OTHER_TXT:
            wdesc, ok = ExpertWidgetChooserDlg.getDialog()
            if ok:
                self.widgetDescription.update(wdesc)
            else:
                return
        else:
            self.widgetDescription['classname'] = choice

        # the name will be set in self.validatePage
        self.wizard().setPanelDescription(PanelDescription('', **self.widgetDescription))
        paneltype = str(self.widgetDescription[
                        'widgetname'] or self.widgetDescription['classname'])
        self.widgetTypeLB.setText("<b>Widget Type:</b> %s" % paneltype)


class AdvSettingsPage(Qt.QWizardPage):

    def __init__(self, parent=None):
        Qt.QWizardPage.__init__(self, parent)

        self.setTitle('Advanced settings')
        self.setSubTitle('Fine-tune the behavior of the panel by assigning a Taurus model ' +
                         'and/or defining the panel interactions with other parts of the GUI')
        self.setFinalPage(True)
        self.models = []

        layout = Qt.QVBoxLayout()

        #----model---------------
        # subwidgets
        self.modelGB = Qt.QGroupBox('Model')
        self.modelGB.setToolTip(
            'Choose a Taurus model to be assigned to the panel')
        # @todo: add a regexp validator (it should return valid on TAURUS_MODEL_LIST_MIME_TYPE)
        self.modelLE = Qt.QLineEdit()
        self.modelChooserBT = Qt.QToolButton()
        self.modelChooserBT.setIcon(Qt.QIcon('designer:devs_tree.png'))
#        self.modelChooser = TaurusModelChooser()

        # connections
        self.modelChooserBT.clicked.connect(self.showModelChooser)
        self.modelLE.editingFinished.connect(self.onModelEdited)

        # layout
        layout1 = Qt.QHBoxLayout()
        layout1.addWidget(self.modelLE)
        layout1.addWidget(self.modelChooserBT)
        self.modelGB.setLayout(layout1)

        #----communications------
        # subwidgets
        self.commGB = Qt.QGroupBox('Communication')
        self.commGB.setToolTip(
            'Define how the panel communicates with other panels and the GUI')
        self.commLV = Qt.QTableView()
        self.commModel = CommTableModel()
        self.commLV.setModel(self.commModel)
        self.commLV.setEditTriggers(self.commLV.AllEditTriggers)
        self.selectedComm = self.commLV.selectionModel().currentIndex()
        self.addBT = Qt.QToolButton()
        self.addBT.setIcon(Qt.QIcon.fromTheme('list-add'))
        self.removeBT = Qt.QToolButton()
        self.removeBT.setIcon(Qt.QIcon.fromTheme('list-remove'))
        self.removeBT.setEnabled(False)

        # layout
        layout2 = Qt.QVBoxLayout()
        layout3 = Qt.QHBoxLayout()
        layout2.addWidget(self.commLV)
        layout3.addWidget(self.addBT)
        layout3.addWidget(self.removeBT)
        layout2.addLayout(layout3)
        self.commGB.setLayout(layout2)

        # connections
        self.addBT.clicked.connect(self.commModel.insertRows)
        self.removeBT.clicked.connect(self.onRemoveRows)
        self.commLV.selectionModel().currentRowChanged.connect(self.onCommRowSelectionChanged)

        layout.addWidget(self.modelGB)
        layout.addWidget(self.commGB)
        self.setLayout(layout)

    def initializePage(self):
        try:
            widget = self.wizard().getPanelDescription().getWidget()
        except Exception as e:
            Logger().debug(repr(e))
            widget = None
        # prevent the user from changing the model if it was already set
        if isinstance(widget, TaurusBaseComponent) and widget.getModelName() != '':
            self.modelLE.setText("(already set by the chosen widget)")
            self.modelGB.setEnabled(False)
        # try to get the SDM as if we were in a TaurusGui app
        try:
            if isinstance(Qt.qApp.SDM, SharedDataManager):
                sdm = Qt.qApp.SDM
        except Exception as e:
            Logger().debug(repr(e))
            sdm = None
        #@todo set selection filter in modelChooser based on the widget's modelclass
        self.itemDelegate = CommItemDelegate(widget=widget, sdm=sdm)
        self.commLV.setItemDelegate(self.itemDelegate)

    def showModelChooser(self):
        models, ok = TaurusModelChooser.modelChooserDlg(
            parent=self, asMimeData=True)
        if not ok:
            return
        self.models = str(models.data(TAURUS_MODEL_LIST_MIME_TYPE))
        self.modelLE.setText(models.text())

    def onModelEdited(self):
        self.models = str(self.modelLE.text())

    def onRemoveRows(self):
        if self.selectedComm.isValid():
            self.commModel.removeRows(self.selectedComm.row())

    def onCommRowSelectionChanged(self, current, previous):
        self.selectedComm = current
        enable = (current.isValid() and 0 <= current.row()
                  < self.commModel.rowCount())
        self.removeBT.setEnabled(enable)

    def validatePage(self):
        desc = self.wizard().getPanelDescription()
        # model
        desc.model = self.models
        # communications
        for uid, slotname, signalname in self.commModel.dumpData():
            if slotname:
                desc.sharedDataRead[uid] = slotname
            if signalname:
                desc.sharedDataWrite[uid] = signalname
        self.wizard().setPanelDescription(desc)
        return True


class CommTableModel(Qt.QAbstractTableModel):
    NUMCOLS = 3
    UID, R, W = list(range(NUMCOLS))

    dataChanged = Qt.pyqtSignal(int, int)

    def __init__(self, parent=None):
        Qt.QAbstractTableModel.__init__(self, parent)
        self.__table = []

    def dumpData(self):
        return copy.deepcopy(self.__table)

    def rowCount(self, index=Qt.QModelIndex()):
        return len(self.__table)

    def columnCount(self, index=Qt.QModelIndex()):
        return self.NUMCOLS

    def headerData(self, section, orientation, role=Qt.Qt.DisplayRole):
        if role == Qt.Qt.TextAlignmentRole:
            if orientation == Qt.Qt.Horizontal:
                return int(Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter)
            return int(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter)
        if role != Qt.Qt.DisplayRole:
            return None
        # So this is DisplayRole...
        if orientation == Qt.Qt.Horizontal:
            if section == self.UID:
                return "Data UID"
            elif section == self.R:
                return "Reader (slot)"
            elif section == self.W:
                return "Writer (signal)"
            return None
        else:
            return str('%i' % (section + 1))

    def data(self, index, role=Qt.Qt.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < self.rowCount()):
            return None
        row = index.row()
        column = index.column()
        # Display Role
        if role == Qt.Qt.DisplayRole:
            text = self.__table[row][column]
            if text == '':
                if column == self.UID:
                    text = '(enter UID)'
                else:
                    text = '(not registered)'
            return str(text)
        return None

    def flags(self, index):
        return (Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled | Qt.Qt.ItemIsDropEnabled | Qt.Qt.ItemIsSelectable)

    def setData(self, index, value=None, role=Qt.Qt.EditRole):
        if index.isValid() and (0 <= index.row() < self.rowCount()):
            row = index.row()
            column = index.column()
            self.__table[row][column] = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def insertRows(self, position=None, rows=1, parentindex=None):
        if position is None:
            position = self.rowCount()
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginInsertRows(parentindex, position, position + rows - 1)
        slice = [self.rowModel() for i in range(rows)]
        self.__table = self.__table[:position] + \
            slice + self.__table[position:]
        self.endInsertRows()
        return True

    def removeRows(self, position, rows=1, parentindex=None):
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginResetModel()
        self.beginRemoveRows(parentindex, position, position + rows - 1)
        self.__table = self.__table[:position] + self.__table[position + rows:]
        self.endRemoveRows()
        self.endResetModel()
        return True

    @staticmethod
    def rowModel(uid='', slot='', signal=''):
        return [uid, slot, signal]


class CommItemDelegate(Qt.QStyledItemDelegate):
    NUMCOLS = 3
    UID, R, W = list(range(NUMCOLS))

    def __init__(self, parent=None, widget=None, sdm=None):
        super(CommItemDelegate, self).__init__(parent)
        if widget is not None:
            widget = weakref.proxy(widget)
        self._widget = widget
        if sdm is not None:
            sdm = weakref.proxy(sdm)
        self._sdm = sdm

    def createEditor(self, parent, option, index):
        column = index.column()
        combobox = Qt.QComboBox(parent)
        combobox.setEditable(True)
        if column == self.UID and self._sdm is not None:
            combobox.addItems(self._sdm.activeDataUIDs())
        elif column == self.R and self._widget is not None:
            slotnames = [n for n, o in inspect.getmembers(
                self._widget, inspect.ismethod) if not n.startswith('_')]
            combobox.addItems(slotnames)
#        elif column==self.W:
#            if self._widget is not None:
#                #@todo: if/when support for old signals style (Qt<4.6) is dropped, we can inspect the methods in search of (new style) signals
#                combobox.addItems(['(Not registered)'])
        return combobox

    def setEditorData(self, editor, index):
        editor.setEditText('')

    def setModelData(self, editor, model, index):
        model.setData(index, editor.currentText())


class PanelDescriptionWizard(Qt.QWizard, TaurusBaseWidget):
    '''A wizard-style dialog for configuring a new TaurusGui panel.
    Use :meth:`getDialog` for launching it
    '''

    def __init__(self, parent=None, designMode=False, gui=None, extraWidgets=None):
        Qt.QWizard.__init__(self, parent)
        name = "PanelDescriptionWizard"
        TaurusBaseWidget.__init__(self, name)
        self._panelDescription = None
        if gui is None:
            gui = parent
        if gui is not None:
            self._gui = weakref.proxy(gui)
        ###self.setOption(self.HaveFinishButtonOnEarlyPages, True)
        #self.namePG = NamePage()
        self.widgetPG = WidgetPage(extraWidgets=extraWidgets)
        self.advSettingsPG = AdvSettingsPage()

        # self.addPage(self.namePG)
        self.addPage(self.widgetPG)
        self.addPage(self.advSettingsPG)

    def getGui(self):
        '''returns a reference to the GUI to which the dialog is associated'''
        return self._gui

    def getPanelDescription(self):
        '''Returns the panel description with the choices made so far

        :return: (PanelDescription) the panel description
        '''
        return self._panelDescription

    def setPanelDescription(self, desc):
        '''Sets the Panel description

        :param desc: (PanelDescription)
        '''
        self._panelDescription = desc

    @staticmethod
    def getDialog(parent, extraWidgets=None):
        """Static method for launching a new Dialog.

        :param parent: parent widget for the new dialog

        :return: (tuple<PanelDescription,bool>) tuple of a description object
                 and a state flag. The state is True if the dialog was accepted
                 and False otherwise
        """
        dlg = PanelDescriptionWizard(parent, extraWidgets=extraWidgets)
        dlg.exec_()
        return dlg.getPanelDescription(), (dlg.result() == dlg.Accepted)


#------------------------------------------------------------------------------

def test():
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(sys.argv)
    form = PanelDescriptionWizard()

    def kk(d):
        print(d)
    Qt.qApp.SDM = SharedDataManager(form)
    Qt.qApp.SDM.connectReader('111111', kk)
    Qt.qApp.SDM.connectWriter('222222', form, 'thisisasignalname')

    form.show()
    sys.exit(app.exec_())


def test2():
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(sys.argv)
    print(ExpertWidgetChooserDlg.getDialog())
    sys.exit()


def main():
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(sys.argv)
    from taurus.qt.qtgui.container import TaurusMainWindow
    form = Qt.QMainWindow()

    def kk(d):
        print(d)
    Qt.qApp.SDM = SharedDataManager(form)
    Qt.qApp.SDM.connectReader('someUID', kk)
    Qt.qApp.SDM.connectWriter('anotherUID', form, 'thisisasignalname')

    form.show()

    paneldesc, ok = PanelDescriptionWizard.getDialog(form, extraWidgets=[('PyQt4.Qt.QLineEdit', 'logos:taurus.png'),
                                                                         ('PyQt4.Qt.QTextEdit', None)])
    if ok:
        w = paneldesc.getWidget(sdm=Qt.qApp.SDM)
        form.setCentralWidget(w)
        form.setWindowTitle(paneldesc.name)
    print(Qt.qApp.SDM.info())

    sys.exit(app.exec_())


if __name__ == "__main__":
    # test2()
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides base widget that displays a Qt view widget envolved
by a toolbar and a status bar (optional)"""

__all__ = ["QBaseModelWidget", "TaurusBaseModelWidget",
           "BaseToolBar", "FilterToolBar", "EditorToolBar", "SelectionToolBar",
           "RefreshToolBar", "PerspectiveToolBar"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt

from taurus.qt.qtgui.util import ActionFactory
from taurus.qt.qtgui.base import TaurusBaseWidget


class BaseToolBar(Qt.QToolBar):

    def __init__(self, name=None, view=None, parent=None, designMode=False):
        if name is None:
            name = "Base toolbar"
        self._viewWidget = view or parent
        Qt.QToolBar.__init__(self, name, parent)
        self.setIconSize(Qt.QSize(16, 16))
        self.setFloatable(False)
        self.setMovable(False)

    def viewWidget(self):
        return self._viewWidget


class FilterToolBar(BaseToolBar):
    """Internal widget providing quick filter to be placed in a _QToolArea"""

    filterEdited = Qt.pyqtSignal('const QString &')
    filterChanged = Qt.pyqtSignal('const QString &')
    clearFilterTriggered = Qt.pyqtSignal()

    def __init__(self, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus filter toolbar", view=view,
                             parent=parent, designMode=designMode)
        filterLineEdit = self._filterLineEdit = Qt.QLineEdit(self)
        filterLineEdit.setSizePolicy(Qt.QSizePolicy(Qt.QSizePolicy.Preferred,
                                                    Qt.QSizePolicy.Preferred))
        filterLineEdit.setToolTip("Quick filter")
        filterLineEdit.textChanged.connect(self.onFilterChanged)
        filterLineEdit.textEdited.connect(self.onFilterEdited)
        self.addWidget(filterLineEdit)

        af = ActionFactory()
        self._clearFilterAction = af.createAction(self, "Clear",
                                                  icon=Qt.QIcon.fromTheme(
                                                      "edit-clear"),
                                                  tip="Clears the filter",
                                                  triggered=self.onClearFilter)
        self.addAction(self._clearFilterAction)

    def getFilterLineEdit(self):
        return self._filterLineEdit

    def onClearFilter(self):
        self.getFilterLineEdit().setText("")
        self.clearFilterTriggered.emit()

    def onFilterChanged(self, text=None):
        text = text or self.getFilterLineEdit().text()
        self.filterChanged.emit(text)

    def onFilterEdited(self, text=None):
        text = text or self.getFilterLineEdit().text()
        self.filterEdited.emit(text)

    def setFilterText(self, text):
        self.getFilterLineEdit().setText(text)


class EditorToolBar(BaseToolBar):
    """Internal widget to be placed in a _QToolArea providing buttons for
    moving, adding and removing items from a view based widget"""

    addTriggered = Qt.pyqtSignal()
    removeTriggered  = Qt.pyqtSignal()
    moveTopTriggered = Qt.pyqtSignal()
    moveDownTriggered = Qt.pyqtSignal()
    moveUpTriggered = Qt.pyqtSignal()
    moveBottomTriggered = Qt.pyqtSignal()

    def __init__(self, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus editor toolbar", view=view,
                             parent=parent, designMode=designMode)

        af = ActionFactory()
        self._addAction = af.createAction(self, "New item",
                                          icon=Qt.QIcon.fromTheme("list-add"),
                                          tip="Add new item",
                                          triggered=self.onAdd)
        self._removeAction = af.createAction(self, "Remove item",
                                             icon=Qt.QIcon.fromTheme(
                                                 "list-remove"),
                                             tip="Remove item",
                                             triggered=self.onRemove)
        self._moveTopAction = af.createAction(self, "To top",
                                              icon=Qt.QIcon.fromTheme("go-top"),
                                              tip="Move selected item to top",
                                              triggered=self.onMoveTop)
        self._moveUpAction = af.createAction(self, "Move up",
                                             icon=Qt.QIcon.fromTheme("go-up"),
                                             tip="Move selected item up one level",
                                             triggered=self.onMoveUp)
        self._moveDownAction = af.createAction(self, "Move down",
                                               icon=Qt.QIcon.fromTheme(
                                                   "go-down"),
                                               tip="Move selected item down one level",
                                               triggered=self.onMoveDown)
        self._moveBottomAction = af.createAction(self, "To bottom",
                                                 icon=Qt.QIcon.fromTheme(
                                                     "go-bottom"),
                                                 tip="Move selected item to bottom",
                                                 triggered=self.onMoveBottom)
        self.addAction(self._addAction)
        self.addAction(self._removeAction)
        self.addAction(self._moveTopAction)
        self.addAction(self._moveUpAction)
        self.addAction(self._moveDownAction)
        self.addAction(self._moveBottomAction)
        #self.setStyleSheet("QWidget {background : red; }")

    def onAdd(self):
        self.addTriggered.emit()

    def onRemove(self):
        self.removeTriggered.emit()

    def onMoveTop(self):
        self.moveTopTriggered.emit()

    def onMoveUp(self):
        self.moveUpTriggered.emit()

    def onMoveDown(self):
        self.moveDownTriggered.emit()

    def onMoveBottom(self):
        self.moveBottomTriggered.emit()


class SelectionToolBar(BaseToolBar):

    selectAllTriggered = Qt.pyqtSignal()
    clearSelectionTriggered = Qt.pyqtSignal()

    def __init__(self, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus selection toolbar", view=view,
                             parent=parent, designMode=designMode)

        af = ActionFactory()
        self._selectAllAction = af.createAction(self, "Select All",
                                                icon=Qt.QIcon.fromTheme(
                                                    "edit-select-all"),
                                                tip="Select all items",
                                                triggered=self.onSelectAll)
        self._clearSelectionAction = af.createAction(self, "Clear selection",
                                                     icon=Qt.QIcon.fromTheme(
                                                         "edit-clear"),
                                                     tip="Clears current selection",
                                                     triggered=self.onclearSelection)

        self.addAction(self._selectAllAction)
        self.addAction(self._clearSelectionAction)

    def onSelectAll(self):
        self.selectAllTriggered.emit()

    def onclearSelection(self):
        self.clearSelectionTriggered.emit()


class RefreshToolBar(BaseToolBar):

    refreshTriggered = Qt.pyqtSignal()

    def __init__(self, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus refresh toolbar", view=view,
                             parent=parent, designMode=designMode)

        af = ActionFactory()
        self._refreshAction = af.createAction(self, "Refresh",
                                              icon=Qt.QIcon.fromTheme(
                                                  "view-refresh"),
                                              tip="Refresh view",
                                              triggered=self.onRefresh)
        self.addAction(self._refreshAction)

    def onRefresh(self):
        self.refreshTriggered.emit()


class PerspectiveToolBar(BaseToolBar):

    perspectiveChanged = Qt.pyqtSignal(str)

    def __init__(self, perspective, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus refresh toolbar", view=view,
                             parent=parent, designMode=designMode)
        self._perspective = perspective
        view = self.viewWidget()
        b = self._perspective_button = Qt.QToolButton(self)
        b.setToolTip("Perspective selection")
        b.setPopupMode(Qt.QToolButton.InstantPopup)
        b.setToolButtonStyle(Qt.Qt.ToolButtonTextBesideIcon)

        menu = Qt.QMenu("Perspective", b)
        b.setMenu(menu)
        af = ActionFactory()
        for persp, persp_data in view.KnownPerspectives.items():
            label = persp_data["label"]
            icon = Qt.QIcon.fromTheme(persp_data["icon"])
            tip = persp_data["tooltip"]
            action = af.createAction(self, label, icon=icon, tip=tip,
                                     triggered=self.onSwitchPerspective)
            action.perspective = persp
            menu.addAction(action)
            if persp == perspective:
                b.setDefaultAction(action)

        self._perspectiveAction = self.addWidget(b)

    def switchPerspectiveButton(self):
        """Returns the QToolButton that handles the switch perspective.

        :return: (PyQt4.QtGui.QToolButton) the switch perspective tool button
        """
        return self._perspective_button

    def onSwitchPerspective(self):
        action = self.sender()
        self._perspective = action.perspective
        self._perspective_button.setDefaultAction(action)
        self.perspectiveChanged.emit(action.perspective)

    def perspective(self):
        return self._perspective


class QBaseModelWidget(Qt.QMainWindow):
    """A pure Qt widget designed to display a Qt view widget (QTreeView for
    example), envolved by optional toolbar and statusbar.
    The Qt model associated with the internal Qt view widget should be a
    :class:`taurus.qt.qtcore.model.TaurusBaseModel`"""

    KnownPerspectives = {}
    DftPerspective = None

    itemClicked = Qt.pyqtSignal(object, int)
    itemDoubleClicked = Qt.pyqtSignal(object, int)
    itemSelectionChanged = Qt.pyqtSignal()
    currentItemChanged = Qt.pyqtSignal(object, object)

    def __init__(self, parent=None, designMode=False, with_filter_widget=True,
                 with_selection_widget=True, with_refresh_widget=True,
                 perspective=None, proxy=None):
        Qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(Qt.Qt.Widget)
        self._baseQModel = None
        self._toolBars = []

        if with_filter_widget:
            if isinstance(with_filter_widget, (bool, int)):
                self._with_filter_widget = FilterToolBar
            else:
                self._with_filter_widget = with_filter_widget
        else:
            self._with_filter_widget = None

        if with_selection_widget:
            if isinstance(with_selection_widget, (bool, int)):
                self._with_selection_widget = SelectionToolBar
            else:
                self._with_selection_widget = with_selection_widget
        else:
            self._with_selection_widget = None

        if with_refresh_widget:
            if isinstance(with_refresh_widget, (bool, int)):
                self._with_refresh_widget = RefreshToolBar
            else:
                self._with_refresh_widget = with_refresh_widget
        else:
            self._with_refresh_widget = None

        self._proxyModel = proxy

        toolBars = self.createToolArea()
        self._viewWidget = self.createViewWidget()
        statusbar = self.createStatusBar()

        for toolBar in toolBars:
            # toolBar.addSeparator()
            self.addToolBar(toolBar)
        self.setContentsMargins(0, 0, 0, 0)
        self.setCentralWidget(self._viewWidget)
        self.setStatusBar(statusbar)

        if perspective is None:
            perspective = self.DftPerspective

        if len(self.KnownPerspectives) > 1:
            p_bar = self._perspectiveBar = PerspectiveToolBar(
                perspective, view=self, parent=self)
            p_bar.perspectiveChanged.connect(self.onSwitchPerspective)
            self.addToolBar(p_bar)
        else:
            self._perspectiveBar = None
        self._setPerspective(perspective)

    def createViewWidget(self, klass=None):
        raise NotImplementedError

    def createStatusBar(self):
        sb = Qt.QStatusBar()
        sb.setSizeGripEnabled(False)
        return sb

    def createToolArea(self):
        tb = []  # tb = self._toolArea = QToolArea(self)
        if self._with_filter_widget:
            f_bar = self._filterBar = self._with_filter_widget(
                view=self, parent=self)
            f_bar.filterChanged.connect(self.onFilterChanged)
            tb.append(f_bar)
        else:
            self._filterBar = None

        if self._with_selection_widget:
            s_bar = self._selectionBar = self._with_selection_widget(
                view=self, parent=self)
            s_bar.selectAllTriggered.connect(self.onSelectAll)
            s_bar.clearSelectionTriggered.connect(self.onClearSelection)
            tb.append(s_bar)
        else:
            self._selectionBar = None

        if self._with_refresh_widget:
            r_bar = self._refreshBar = self._with_refresh_widget(
                view=self, parent=self)
            r_bar.refreshTriggered.connect(self.onRefreshModel)
            tb.append(r_bar)
        else:
            self._refreshBar = None

        return tb

    def getPerspectiveBar(self):
        return self._perspectiveBar

    def getFilterBar(self):
        return self._filterBar

    def getSelectionBar(self):
        return self._selectionBar

    def getRefreshBar(self):
        return self._refreshBar

    def onRefreshModel(self):
        self.getQModel().refresh(True)

    def onSelectAll(self):
        view = self.viewWidget()
        view.selectAll()

    def onClearSelection(self):
        view = self.viewWidget()
        view.clearSelection()

    def _onClicked(self, index):
        '''Emits an "itemClicked" signal with with the clicked item and column
        as arguments'''
        item = self._mapToSource(index).internalPointer()
        self.itemClicked.emit(item, index.column())

    def _onDoubleClicked(self, index):
        '''Emits an "itemDoubleClicked" signal with the clicked item and column
        as arguments'''
        item = self._mapToSource(index).internalPointer()
        self.itemDoubleClicked.emit(item, index.column())

    def viewWidget(self):
        return self._viewWidget

    def getQModel(self):
        return self.viewWidget().model()

    def getBaseQModel(self):
        return self._baseQModel

    def usesProxyQModel(self):
        return isinstance(self.getQModel(), Qt.QAbstractProxyModel)

    def _mapToSource(self, index):
        if not self.usesProxyQModel():
            return index
        model = self.getQModel()
        while isinstance(model, Qt.QAbstractProxyModel):
            index = model.mapToSource(index)
            model = model.sourceModel()
        return index

    def setQModel(self, qmodel):

        self._baseQModel = qmodel
        while isinstance(self._baseQModel, Qt.QAbstractProxyModel):
            self._baseQModel = self._baseQModel.sourceModel()

        view = self.viewWidget()
        old_selection_model = view.selectionModel()
        CC = 'currentChanged(const QModelIndex &,const QModelIndex &)'
        SC = 'selectionChanged(QItemSelection &, QItemSelection &)'
        if old_selection_model is not None:
            old_selection_model.currentChanged.disconnect(
                                  self.viewCurrentIndexChanged)
            old_selection_model.selectionChanged.disconnect(
                                  self.viewSelectionChanged)
        view.setModel(qmodel)
        new_selection_model = view.selectionModel()
        if new_selection_model is not None:
            new_selection_model.currentChanged.connect(
                               self.viewCurrentIndexChanged)
            new_selection_model.selectionChanged.connect(
                               self.viewSelectionChanged)
        view.setCurrentIndex(view.rootIndex())
        self._updateToolBar()

    def viewSelectionChanged(self, selected, deselected):
        self.itemSelectionChanged.emit()

    def viewCurrentIndexChanged(self, current, previous):
        # if there is a proxy model we have to translate the selection
        base_current = self._mapToSource(current)
        base_previous = self._mapToSource(previous)

        self._updateToolBar(current)

        if base_current.isValid():
            currentTaurusTreeItem = base_current.internalPointer()
        else:
            currentTaurusTreeItem = None

        if base_previous.isValid():
            previousTaurusTreeItem = base_previous.internalPointer()
        else:
            previousTaurusTreeItem = None
        self.currentItemChanged.emit(currentTaurusTreeItem,
                  previousTaurusTreeItem)

    def _updateToolBar(self, current=None, previous=None):
        pass

    def selectedItems(self):
        """Returns a list of all selected non-hidden items

        :return: (list<TaurusTreeItem>)
        """
        view = self.viewWidget()
        return [self._mapToSource(index).internalPointer() for index in view.selectedIndexes()]

    def onFilterChanged(self, filter):
        if not self.usesProxyQModel():
            return
        proxy_model = self.getQModel()
        if len(filter) > 0 and filter[0] != '^':
            filter = '^' + filter
        proxy_model.setFilterRegExp(filter)
        # proxy_model.setFilterFixedString(filter)
        # proxy_model.setFilterWildcard(filter)
        # self.update()

    def refresh(self):
        self.getQModel().refresh()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Perspective handling
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def perspective(self):
        return self._perspectiveBar.perspective()

    def onSwitchPerspective(self, perspective):
        old_qmodel = self.getQModel()
        self._setPerspective(perspective)

        # set the selectables as they where in the previous model
        if hasattr(old_qmodel, "selectables"):
            self.getQModel().setSelectables(old_qmodel.selectables())

        # set the taurus model (if any) to the qmodel
        if hasattr(self, 'getModelObj'):
            taurusModel = self.getModelObj()
            if taurusModel is not None:
                self.getQModel().setDataSource(taurusModel)

    def _setPerspective(self, perspective):
        qmodel_classes = self.KnownPerspectives[perspective]["model"]
        # reversed
        qmodel_class, qmodel_proxy_classes = qmodel_classes[
            -1], qmodel_classes[-2::-1]
        qmodel = qmodel_class(parent=self)
        qmodel_source = qmodel
        if self._proxyModel is None:  # applies the chain of proxies
            for qmodel_proxy_class in qmodel_proxy_classes:
                qproxy = qmodel_proxy_class(self)
                qproxy.setSourceModel(qmodel_source)
                qmodel_source = qproxy
        else:
            self._proxyModel.setSourceModel(qmodel_source)
            qmodel_source = self._proxyModel
        self.setQModel(qmodel_source)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QMainWindow overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def addToolBar(self, toolbar):
        Qt.QMainWindow.addToolBar(self, toolbar)
        self._toolBars.append(toolbar)

    def insertToolBar(self, before, toolbar):
        if isinstance(before, Qt.QToolBar):
            index = self._toolBars.index(before)
        else:
            index = before
            before = self._toolBars[before]
        Qt.QMainWindow.insertToolBar(self, before, toolbar)
        self._toolBars.insert(index, toolbar)


class TaurusBaseModelWidget(TaurusBaseWidget):
    """A class:`taurus.qt.qtgui.base.TaurusBaseWidget` that connects to a
    taurus model. It must be used together with
    class:`~taurus.qt.qtgui.base.QBaseModelWidget`"""

    def __init__(self, designMode=False):  # , perspective=None, proxy=None):
        name = self.__class__.__name__
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setModel(self, m):
        TaurusBaseWidget.setModel(self, m)

        view, modelObj = self.viewWidget(), self.getModelObj()
        model = view.model()
        if model is None:
            return
        model.setDataSource(modelObj)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    #: This property holds the unique URI string representing the model name
    #: with which this widget will get its data from. The convention used for
    #: the string can be found :ref:`here <model-concept>`.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusBaseWidget.getModel`
    #:     * :meth:`TaurusBaseWidget.setModel`
    #:     * :meth:`TaurusBaseWidget.resetModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel, setModel,
                            TaurusBaseWidget.resetModel)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides basic pure Qt container widgets"""

from builtins import str

import json

from taurus.external.qt import Qt
from taurus.qt.qtgui.icon import getStandardIcon

__all__ = ["QGroupWidget"]

__docformat__ = 'restructuredtext'


_TitleBarStyleExpanded = """.QFrame {{
border-width: 0px;
border-style: solid;
border-color: {stop_color};
border-top-left-radius: {border_radius};
border-top-right-radius: {border_radius};
border-bottom-left-radius: 0px;
border-bottom-right-radius: 0px;
background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                  stop: 0 {start_color}, stop: 1 {stop_color});
}}"""

_TitleBarStyleCollapsed = """.QFrame {{
border-width: 0px;
border-style: solid;
border-color: {stop_color};
border-top-left-radius: {border_radius};
border-top-right-radius: {border_radius};
border-bottom-left-radius: {border_radius};
border-bottom-right-radius: {border_radius};
background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                  stop: 0 {start_color}, stop: 1 {stop_color});
}}"""

_TitleLabelStyle = """.QLabel {{ color : {font_color}; }}"""

_ContentBarStyleWithTitle = """ContentFrame {{
border-top-width: 0px;
border-left-width: 1px;
border-right-width: 1px;
border-bottom-width: 1px;
border-style: solid;
border-color: {border_color};
border-top-left-radius: 0px;
border-top-right-radius: 0px;
border-bottom-left-radius: {border_radius};
border-bottom-right-radius: {border_radius};
background-color: qlineargradient(x1: 0, y1: 0, x2: 1.0, y2: 1.0,
                                  stop: 0 {start_color}, stop: 1 {stop_color});
/*
  background-position: center center;
*/
}}"""

_ContentBarStyleWithoutTitle = """ContentFrame {{
border-width: 1px;
border-style: solid;
border-color: {border_color};
border-top-left-radius: {border_radius};
border-top-right-radius: {border_radius};
border-bottom-left-radius: {border_radius};
border-bottom-right-radius: {border_radius};
background-color: qlineargradient(x1: 0, y1: 0, x2: 1.0, y2: 1.0,
                                  stop: 0 {start_color}, stop: 1 {stop_color});
/*
  background-position: center center;
*/
}}"""

# Empty content QFrame to avoid frame style to be propagated
# to child QFrame widgets


class ContentFrame(Qt.QFrame):
    pass


class QGroupWidget(Qt.QWidget):
    """An expandable/collapsible composite widget"""

    DefaultTitleBarVisible = True
    DefaultTitleBarHeight = 16
    DefaultTitleBarStyle = {
        'start_color': 'rgb(60, 150, 255)',
        'stop_color': 'rgb(0, 65, 200)',
        'font_color': 'white',
        'border_radius': '5px',
    }

    DefaultContentVisible = True
    DefaultContentStyle = {
        'start_color': 'rgb(224, 224, 224)',
        'stop_color': 'rgb(255, 255, 255)',
        'border_color': 'rgb(0, 85, 227)',
        'border_radius': '5px',
    }

    def __init__(self, parent=None, designMode=False):
        Qt.QWidget.__init__(self, parent)
        self._titleVisible = self.DefaultTitleBarVisible
        self._contentVisible = self.DefaultContentVisible
        self._titleBarStyle = self.DefaultTitleBarStyle
        self._contentStyle = self.DefaultContentStyle
        self.__init()
        self._updateStyle()
        self.resetContentVisible()
        self.resetTitleHeight()
        self.resetTitleVisible()

    def __init(self):
        panelLayout = Qt.QVBoxLayout()
        panelLayout.setSpacing(0)
        panelLayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(panelLayout)

        self._titleBar = titleBar = Qt.QFrame()
        panelLayout.addWidget(titleBar, 0)

        l = Qt.QHBoxLayout()
        l.setContentsMargins(2, 2, 2, 2)
        l.setSpacing(2)
        self._titleBar.setLayout(l)

        self._titleButton = Qt.QToolButton()
        self._titleButton.setStyleSheet("border: 0px")
        styleOption = Qt.QStyleOption()
        styleOption.initFrom(self._titleButton)
        style = Qt.QApplication.style()
        icon = style.standardIcon(
            Qt.QStyle.SP_DesktopIcon, styleOption, self._titleButton)
        self._titleButton.setIcon(icon)
        self._titleLabel = Qt.QLabel()
        self._upDownButton = Qt.QToolButton()
        self._upDownButton.setStyleSheet("border: 0px")
        self._upDownButton.clicked.connect(self.switchContentVisible)
        l.addWidget(self._titleButton, 0)
        l.addWidget(self._titleLabel, 1)
        l.addWidget(self._upDownButton, 0)

        self._content = content = ContentFrame()
        l = Qt.QHBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)
        content.setLayout(l)
        panelLayout.addWidget(content, 1)

    def _updateStyle(self):
        """Internal method that updates the style """
        if self.contentVisible:
            ts = _TitleBarStyleExpanded
        else:
            ts = _TitleBarStyleCollapsed
        fullTitleBarStyle = ts.format(**self._titleBarStyle)
        fullTitleLabelStyle = _TitleLabelStyle.format(**self._titleBarStyle)
        if self.titleVisible:
            contentStyleTemplate = _ContentBarStyleWithTitle
        else:
            contentStyleTemplate = _ContentBarStyleWithoutTitle

        contentStyle = self._contentStyle.copy()
        contentStyle['border_color'] = self._titleBarStyle['stop_color']
        fullContentStyle = contentStyleTemplate.format(**contentStyle)
        self._titleBar.setStyleSheet(fullTitleBarStyle)
        self._titleLabel.setStyleSheet(fullTitleLabelStyle)
        self._content.setStyleSheet(fullContentStyle)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {'module': 'taurus.qt.qtgui.container',
                'group': 'Taurus Containers',
                'icon': "designer:groupwidget.png",
                'container': True}

    def content(self):
        """Returns the contents widget

        :return: (Qt.QFrame) the content widget"""
        return self._content

    def titleBar(self):
        """Returns the title bar widget

        :return: (Qt.QFrame) the title bar widget"""
        return self._titleBar

    def titleButton(self):
        """Returns the title button widget

        :return: (Qt.QToolButton) the title button widget"""
        return self._titleButton

    def collapseButton(self):
        """Returns the collapse button widget

        :return: (Qt.QToolButton) the collapse button widget"""
        return self._upDownButton

    def setTitle(self, title):
        """Sets this widget's title

        :param title: (str) the new widget title"""
        self._titleLabel.setText(title)

    def getTitle(self):
        """Returns this widget's title

        :return: (str) this widget's title"""
        return self._titleLabel.text()

    def setTitleIcon(self, icon):
        """Sets this widget's title icon

        :param icon: (Qt.QIcon) the new widget title icon"""
        self._titleButton.setIcon(icon)

    def getTitleIcon(self):
        """Returns this widget's title icon

        :return: (Qt.QIcon) this widget's title icon"""
        return self._titleButton.icon()

    def switchContentVisible(self):
        """Switches this widget's contents visibility"""
        self.setContentVisible(not self.isContentVisible())

    def isContentVisible(self):
        """Returns this widget's contents visibility

        :return: (bool) this widget's contents visibility"""
        return self._contentVisible

    def resetContentVisible(self):
        """Resets this widget's contents visibility"""
        self.setContentVisible(self.DefaultContentVisible)

    def setContentVisible(self, show):
        """Sets this widget's contents visibility

        :param show: (bool) the new widget contents visibility"""
        self._contentVisible = show
        self._updateStyle()

        #if show: icon_name = "go-previous"
        # else: icon_name = "go-down"
        #icon = Qt.QIcon.fromTheme(icon_name)

        if show:
            icon_name = Qt.QStyle.SP_TitleBarShadeButton
        else:
            icon_name = Qt.QStyle.SP_TitleBarUnshadeButton
        icon = getStandardIcon(icon_name, self._upDownButton)

        self._upDownButton.setIcon(icon)
        self._content.setVisible(show)
        self.adjustSize()

    def isTitleVisible(self):
        """Returns this widget's title visibility

        :return: (bool) this widget's title visibility"""
        return self._titleVisible

    def resetTitleVisible(self):
        """Resets this widget's title visibility"""
        self.setTitleVisible(self.DefaultTitleBarVisible)

    def setTitleVisible(self, show):
        """Sets this widget's title visibility

        :param icon: (bool) the new widget title visibility"""
        self._titleVisible = show
        self._titleBar.setVisible(show)
        self._updateStyle()

    def getTitleHeight(self):
        """Returns this widget's title height

        :return: (bool) this widget's title height"""
        return self.titleButton().iconSize().height()

    def setTitleHeight(self, h):
        """Sets this widget's title height

        :param icon: (bool) the new widget title height"""
        s = Qt.QSize(h, h)
        self.titleButton().setIconSize(s)
        self.collapseButton().setIconSize(s)

    def resetTitleHeight(self):
        """Resets this widget's title height"""
        self.setTitleHeight(self.DefaultTitleBarHeight)

    def getTitleStyle(self):
        """Returns this widget's title style

        :return: (dict) this widget's title style"""
        return self._titleBarStyle

    def setTitleStyle(self, style_map):
        """Sets this widget's title style
        Used key/values for style_map:
        - 'start_color'  : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'stop_color'   : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'font_color'   : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'border_radius': radius (Ex.: '5px', '5px,2px')

        :param style_map: (dict) the new widget title style"""
        style = self.DefaultTitleBarStyle.copy()
        style.update(style_map)
        self._titleBarStyle = style
        self._updateStyle()

    def resetTitleStyle(self):
        """Resets this widget's title style"""
        self.setTitleStyle({})

    def getTitleStyleStr(self):
        """Returns this widget's title style

        :return: (dict) this widget's title style"""
        return json.dumps(self._titleBarStyle)

    def setTitleStyleStr(self, style_map):
        """Sets this widget's title style
        Used key/values for style_map:
        - 'start_color'  : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'stop_color'   : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'font_color'   : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'border_radius': radius (Ex.: '5px', '5px,2px')

        :param style_map: (dict) the new widget title style"""
        style_map = json.loads(str(style_map))
        self.setTitleStyle(style_map)

    def resetTitleStyleStr(self):
        """Resets this widget's title style"""
        self.resetTitleStyle()

    def getContentStyle(self):
        """Returns this widget's content style

        :return: (dict) this widget's content style"""
        return self._contentStyle

    def setContentStyle(self, style_map):
        """Sets this widget's content style
        Used key/values for style_map:
        - 'start_color'  : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'stop_color'   : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')

        :param style_map: (dict) the new widget content style"""
        style = self.DefaultContentStyle.copy()
        style.update(style_map)
        self._contentStyle = style
        self._updateStyle()

    def resetContentStyle(self):
        """Resets this widget's content style"""
        self.setContentStyle({})

    def getContentStyleStr(self):
        """Returns this widget's content style

        :return: (dict) this widget's content style"""
        return json.dumps(self._contentStyle)

    def setContentStyleStr(self, style_map):
        """Sets this widget's content style
        Used key/values for style_map:
        - 'start_color'  : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')
        - 'stop_color'   : brush (Ex.: '#E0E0E0', 'rgb(0,0,0)', 'white')

        :param style_map: (dict) the new widget content style"""
        style_map = json.loads(str(style_map))
        self.setContentStyle(style_map)

    def resetContentStyleStr(self):
        """Resets this widget's content style"""
        self.resetContentStyle()

    #: This property contains the widget's title
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.getTitle`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setTitle`
    title = Qt.pyqtProperty("QString", getTitle, setTitle)

    #: This property contains the widget's title icon
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.getTitleIcon`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setTitleIcon`
    titleIcon = Qt.pyqtProperty("QIcon", getTitleIcon, setTitleIcon)

    #: This property contains the widget's title height
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.getTitleHeight`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setTitleHeight`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.resetTitleHeight`
    titleHeight = Qt.pyqtProperty(
        "int", getTitleHeight, setTitleHeight, resetTitleHeight)

    #: This property contains the widget's title visibility
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.isTitleVisible`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setTitleVisible`
    titleVisible = Qt.pyqtProperty("bool", isTitleVisible, setTitleVisible)

    #: This property contains the widget's content's visibility
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.isContentVisible`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setContentVisible`
    #:     * :meth:`taurus.qt.qtgui.container.QGroupWidget.resetContentVisible`
    contentVisible = Qt.pyqtProperty(
        "bool", isContentVisible, setContentVisible, resetContentVisible)

    # : This property contains the widget's content style
    # : The style must be a json dictionary
    # :
    # : **Access functions:**
    # :
    # :     * :meth:`taurus.qt.qtgui.container.QGroupWidget.getContentStyleStr`
    # :     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setContentStyleStr`
    # :     * :meth:`taurus.qt.qtgui.container.QGroupWidget.resetContentStyleStr`
    contentStyle = Qt.pyqtProperty("QString", getContentStyleStr, setContentStyleStr,
                                   resetContentStyleStr, doc="The style must be a json dictionary")

    # : This property contains the widget's title style
    # : The style must be a json dictionary
    # :
    # : **Access functions:**
    # :
    # :     * :meth:`taurus.qt.qtgui.container.QGroupWidget.getTitleStyleStr`
    # :     * :meth:`taurus.qt.qtgui.container.QGroupWidget.setTitleStyleStr`
    # :     * :meth:`taurus.qt.qtgui.container.QGroupWidget.resetTitleStyleStr`
    titleStyle = Qt.pyqtProperty("QString", getTitleStyleStr, setTitleStyleStr,
                                 resetTitleStyleStr, doc="The style must be a json dictionary")


def demo():
    "QGroup Widget"
    w = Qt.QWidget()
    l = Qt.QVBoxLayout()
    w.setLayout(l)

    panel = QGroupWidget()
    panel.title = "Database"
    contentLayout = Qt.QFormLayout()
    panel.content().setLayout(contentLayout)
    contentLayout.addRow("&Host", Qt.QLineEdit())
    contentLayout.addRow("&Port", Qt.QLineEdit())
    l.addWidget(panel, 0)

    panel = QGroupWidget()
    panel.title = "Hello world"
    panel.titleIcon = Qt.QIcon.fromTheme("video-x-generic")
    panel.setTitleStyle({
        'start_color': 'rgb(255, 60, 60)',
        'stop_color': 'rgb(200, 0, 0)',
        'font_color': 'rgb(140, 0, 0)',
        'border_radius': '10px',
    })
    panel.setContentStyle({
        'border_radius': '0px',
    })
    contentLayout = Qt.QFormLayout()
    panel.content().setLayout(contentLayout)
    contentLayout.addRow("State", Qt.QPushButton("Press here"))
    contentLayout.addRow("Status", Qt.QLineEdit())
    contentLayout.addRow("Coment", Qt.QLineEdit())
    contentLayout.addRow("Build", Qt.QCheckBox())
    contentLayout.addRow("Upper limit", Qt.QSpinBox())
    contentLayout.addRow("Lower limit", Qt.QSpinBox())
    l.addWidget(panel, 0)

    panel = QGroupWidget()
    panel.title = "Hello world 2"
    panel.titleIcon = Qt.QIcon.fromTheme("network-server")
    panel.titleVisible = False
    contentLayout = Qt.QFormLayout()
    panel.content().setLayout(contentLayout)
    contentLayout.addRow("Something", Qt.QLineEdit())
    contentLayout.addRow("More", Qt.QLineEdit())
    l.addWidget(panel, 0)

    panel = QGroupWidget()
    panel.title = "5"
    panel.titleIcon = Qt.QIcon.fromTheme("folder")
    contentLayout = Qt.QVBoxLayout()
    panel.content().setLayout(contentLayout)
    panel2 = QGroupWidget()
    panel2.title = "5.1"
    panel2.titleIcon = Qt.QIcon.fromTheme("folder")
    panel2.titleHeight = 48
    contentLayout2 = Qt.QFormLayout()
    panel2.content().setLayout(contentLayout2)
    contentLayout2.addRow("Something", Qt.QLineEdit())
    contentLayout2.addRow("More", Qt.QLineEdit())
    contentLayout.addWidget(panel2, 0)
    l.addWidget(panel, 0)

    l.addStretch(1)

    w.show()
    w.adjustSize()
    return w


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="Group widget demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    w = demo()
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides a pint unit registry instance (`UR`) to be used by all
taurus objects. It also provides the `Quantity` factory from that registry
(also aliased as `Q_`).
"""

from pint import UnitRegistry

# Ininitialize the unit registry for taurus
UR = UnitRegistry()
UR.default_format = '~'  # use abbreviated units
Q_ = Quantity = UR.Quantity



# -*- coding: utf-8 -*-
#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains all taurus tango attribute"""

from builtins import range
from builtins import str

import re
import time
import threading
import weakref
import PyTango
import numpy
from functools import partial

from taurus import Manager
from taurus.core.units import Quantity, UR

from taurus import tauruscustomsettings
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.taurusbasetypes import (TaurusEventType,
                                         TaurusSerializationMode,
                                         SubscriptionState, TaurusAttrValue,
                                         DataFormat, DataType)
from taurus.core.taurusoperation import WriteAttrOperation
from taurus.core.util.event import EventListener, _BoundMethodWeakrefWithCall
from taurus.core.util.log import (debug, taurus4_deprecation,
                                  deprecation_decorator)

from taurus.core.tango.enums import (EVENT_TO_POLLING_EXCEPTIONS,
                                     FROM_TANGO_TO_NUMPY_TYPE,
                                     DevState)

from .util.tango_taurus import (description_from_tango,
                                display_level_from_tango,
                                quality_from_tango,
                                standard_display_format_from_tango,
                                quantity_from_tango_str,
                                data_format_from_tango,
                                data_type_from_tango)

__all__ = ["TangoAttribute", "TangoAttributeEventListener", "TangoAttrValue"]

__docformat__ = "restructuredtext"


class TangoAttrValue(TaurusAttrValue):
    """A TaurusAttrValue specialization to decode PyTango.DeviceAttribute
    objects
    """

    def __init__(self, attr=None, pytango_dev_attr=None, config=None):
        # config parameter is kept for backwards compatibility only
        TaurusAttrValue.__init__(self)
        if config is not None:
            from taurus.core.util.log import deprecated
            deprecated(dep='"config" kwarg', alt='"attr"', rel='4.0')
            attr = config
        if attr is None:
            self._attrRef = None
            self.__attrType = None
        else:
            self._attrRef = weakref.proxy(attr)
            self.__attrName = attr.getFullName()
            self.__attrType = attr.type

        self.config = self._attrRef  # bck-compat

        self._pytango_dev_attr = p = pytango_dev_attr
        if p is None:
            self._pytango_dev_attr = p = PyTango.DeviceAttribute()
            return

        if self._attrRef is None:
            return

        numerical = PyTango.is_numerical_type(self._attrRef._tango_data_type,
                                              inc_array=True)

        if p.has_failed:
            self.error = PyTango.DevFailed(*p.get_err_stack())
        else:
            # spectra and images can be empty without failing
            if p.is_empty and self._attrRef.data_format != DataFormat._0D:
                dtype = FROM_TANGO_TO_NUMPY_TYPE.get(
                    self._attrRef._tango_data_type)
                if self._attrRef.data_format == DataFormat._1D:
                    shape = (0,)
                elif self._attrRef.data_format == DataFormat._2D:
                    shape = (0, 0)
                p.value = numpy.empty(shape, dtype=dtype)
                if not (numerical or self._attrRef.type == DataType.Boolean):
                    # generate a nested empty list of given shape
                    p.value = []
                    for _ in range(len(shape) - 1):
                        p.value = [p.value]

        rvalue = p.value
        wvalue = p.w_value
        if numerical:
            units = self._attrRef._units
            if rvalue is not None:
                rvalue = Quantity(rvalue, units=units)
            if wvalue is not None:
                wvalue = Quantity(wvalue, units=units)
        elif isinstance(rvalue, PyTango._PyTango.DevState):
            rvalue = DevState[str(rvalue)]

        self.rvalue = rvalue
        self.wvalue = wvalue
        self.time = p.time  # TODO: decode this into a TaurusTimeVal
        self.quality = quality_from_tango(p.quality)

    def __getattr__(self, name):
        """
        If the member `name` is not defined in this class, try to get it
        from the TangoAttribute (configuration) or from the PyTango value
        """
        # Do not try to delegate special methods
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError("'%s' object has no attribute %s"
                                 % (self.__class__.__name__, name))
        try:
            # maybe name is defined in the TangoAttribute?
            try:
                # Use the attr reference if the attr is still valid
                ret = getattr(self._attrRef, name)
            except ReferenceError:
                # re-create the attribute in case it was no longer referenced
                # (this may happen in some rare cases in which the value is
                # still used but the attr is no longer referenced elsewhere)
                a = TangoAttribute.factory().getAttribute(self.__attrName)
                ret = getattr(a, name)
        except AttributeError:
            try:
                # maybe name is defined in the PyTango value?
                ret = getattr(self._pytango_dev_attr, name)
            except AttributeError:
                raise AttributeError("'%s' object has no attribute %s"
                                     % (self.__class__.__name__, name))
        # return the attr but only after warning
        from taurus.core.util.log import deprecated
        deprecated(dep='TangoAttrValue.%s' % name,
                   alt='TangoAttribute.%s' % name, rel='4.0')
        return ret

    # --------------------------------------------------------
    # This is for backwards compat with the API of taurus < 4
    #
    @taurus4_deprecation(alt='.rvalue')
    def _get_value(self):
        """for backwards compat with taurus < 4"""
        try:
            return self.__fix_int(self.rvalue.magnitude)
        except AttributeError:
            return self.rvalue

    @taurus4_deprecation(alt='.rvalue')
    def _set_value(self, value):
        """for backwards compat with taurus < 4"""
        debug('Setting %r to %s' % (value, self.name))

        if self.rvalue is None:  # we do not have a previous rvalue
            import numpy
            dtype = numpy.array(value).dtype
            if numpy.issubdtype(dtype, int) or numpy.issubdtype(dtype, float):
                msg = 'Refusing to set ambiguous value (deprecated .value API)'
                raise ValueError(msg)
            else:
                self.rvalue = value
        elif hasattr(self.rvalue, 'units'):  # we do have it and is a Quantity
            self.rvalue = Quantity(value, units=self.rvalue.units)
        else:  # we do have a previous value and is not a quantity
            self.rvalue = value

    value = property(_get_value, _set_value)

    @taurus4_deprecation(alt='.wvalue')
    def _get_w_value(self):
        """for backwards compat with taurus < 4"""
        debug(repr(self))
        try:
            return self.__fix_int(self.wvalue.magnitude)
        except AttributeError:
            return self.wvalue

    @taurus4_deprecation(alt='.wvalue')
    def _set_w_value(self, value):
        """for backwards compat with taurus < 4"""
        debug('Setting %r to %s' % (value, self.name))

        if self.wvalue is None:  # we do not have a previous wvalue
            import numpy
            dtype = numpy.array(value).dtype
            if numpy.issubdtype(dtype, int) or numpy.issubdtype(dtype, float):
                msg = 'Refusing to set ambiguous value (deprecated .value API)'
                raise ValueError(msg)
            else:
                self.wvalue = value
        elif hasattr(self.wvalue, 'units'):  # we do have it and is a Quantity
            self.wvalue = Quantity(value, units=self.wvalue.units)
        else:  # we do have a previous value and is not a quantity
            self.wvalue = value

    w_value = property(_get_w_value, _set_w_value)

    @property
    @taurus4_deprecation(alt='.error')
    def has_failed(self):
        return self.error

    def __fix_int(self, value):
        """cast value to int if it is an integer.
        Works on scalar and non-scalar values
        """
        if self.__attrType != DataType.Integer:
            return value
        try:
            return int(value)
        except TypeError:
            import numpy
            return numpy.array(value, dtype='int')


class TangoAttribute(TaurusAttribute):

    no_cfg_value = '-----'
    no_unit = 'No unit'
    no_standard_unit = 'No standard unit'
    no_display_unit = 'No display unit'
    no_description = 'No description'
    not_specified = 'Not specified'
    no_min_value = no_max_value = not_specified
    no_min_alarm = no_max_alarm = not_specified
    no_min_warning = no_max_warning = not_specified
    no_delta_t = no_delta_val = not_specified
    no_rel_change = no_abs_change = not_specified
    no_archive_rel_change = no_archive_abs_change = not_specified
    no_archive_period = not_specified

    # helper class property that stores a reference to the corresponding
    # factory
    _factory = None
    _scheme = 'tango'
    _description = 'A Tango Attribute'

    def __init__(self, name='', parent=None, **kwargs):
        # the last attribute value
        self.__attr_value = None

        # the last attribute error
        self.__attr_err = None

        # Lock for protecting the critical read region 
        # where __attr_value and __attr_err are updated
        self.__read_lock = threading.RLock()

        # the change event identifier
        self.__chg_evt_id = None

        # current event subscription state
        self.__subscription_state = SubscriptionState.Unsubscribed
        self.__subscription_event = threading.Event()

        # the parent's HW object (the PyTango Device obj)
        self.__dev_hw_obj = None

        # unit for which a decode warning has already been issued
        self.__already_warned_unit = None

        self.call__init__(TaurusAttribute, name, parent, **kwargs)

        attr_info = None
        if parent:
            attr_name = self.getSimpleName()
            try:
                attr_info = parent.attribute_query(attr_name)
            except (AttributeError, PyTango.DevFailed):
                # if PyTango could not connect to the dev
                attr_info = None

        # Set default values in case the attrinfoex is None
        self.writable = False
        dis_level = PyTango.DispLevel.OPERATOR
        self.display_level = display_level_from_tango(dis_level)
        self.tango_writable = PyTango.AttrWriteType.READ
        self._units = self._unit_from_tango(PyTango.constants.UnitNotSpec)
        # decode the Tango configuration attribute (adds extra members)
        self._pytango_attrinfoex = None
        self._decodeAttrInfoEx(attr_info)

        # subscribe to configuration events (unsubscription done at cleanup)
        self.__cfg_evt_id = None
        if self.factory().is_tango_subscribe_enabled():
            self._subscribeConfEvents()

    def __del__(self):
        self.cleanUp()

    def cleanUp(self):
        self.trace("[TangoAttribute] cleanUp")
        self._unsubscribeConfEvents()
        TaurusAttribute.cleanUp(self)
        self.__dev_hw_obj = None
        self._pytango_attrinfoex = None

    def __getattr__(self, name):
        try:
            return getattr(self._pytango_attrinfoex, name)
        except AttributeError:
            raise AttributeError("'TangoAttribute' object has no attribute %s"
                                 % name)

    def getNewOperation(self, value):
        attr_value = PyTango.AttributeValue()
        attr_value.name = self.getSimpleName()
        attr_value.value = self.encode(value)
        op = WriteAttrOperation(self, attr_value)
        return op

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # PyTango connection
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def isInteger(self, inc_array=False):
        tgtype = self._tango_data_type
        return PyTango.is_int_type(tgtype, inc_array=inc_array)

    def isFloat(self, inc_array=False):
        tgtype = self._tango_data_type
        return PyTango.is_float_type(tgtype, inc_array=inc_array)

    def isBoolean(self, inc_array=False):
        tgtype = self._tango_data_type
        # PyTango.is_bool_type is not implemented in Tango7 and neither in
        # some Tango8, at least 8.1.1. Avoid to use is_bool_type method
        # while taurus is still compatible with these versions.
        # PyTango.is_bool_type(tgtype, inc_array=inc_array)
        if tgtype == PyTango.CmdArgType.DevBoolean:
            return True
        if inc_array and tgtype == PyTango.CmdArgType.DevVarBooleanArray:
            return True
        return False

    def isState(self):
        """
        returns whether the attribute of tango DevState type
        """
        tgtype = self._tango_data_type
        return tgtype == PyTango.CmdArgType.DevState

    def getFormat(self, cache=True):
        return self.format

    def encode(self, value):
        """Translates the given value into a tango compatible value according to
        the attribute data type.

        Raises `pint.DimensionalityError` if value is a Quantity and it
        cannot be expressed in the units of the attribute set in the DB
        """
        if isinstance(value, Quantity):
            # convert to units of the attr in the DB (or raise an exception)
            magnitude = value.to(self._units).magnitude
        else:
            magnitude = value

        fmt = self.getDataFormat()
        tgtype = self._tango_data_type
        if fmt == DataFormat._0D:
            if tgtype == PyTango.CmdArgType.DevDouble:
                attrvalue = float(magnitude)
            elif tgtype == PyTango.CmdArgType.DevFloat:
                # We encode to float, but rounding to Tango::DevFloat precision
                # see: http://sf.net/p/sardana/tickets/162
                attrvalue = float(numpy.float32(magnitude))
            elif PyTango.is_int_type(tgtype):
                # changed as a partial workaround to a problem in PyTango
                # writing to DevULong64 attributes (see ALBA RT#29793)
                attrvalue = int(magnitude)
            elif tgtype == PyTango.CmdArgType.DevBoolean:
                try:
                    attrvalue = bool(int(magnitude))
                except:
                    attrvalue = str(magnitude).lower() == 'true'
            elif tgtype == PyTango.CmdArgType.DevUChar:
                attrvalue = int(magnitude)
            elif tgtype in (PyTango.CmdArgType.DevState,
                            PyTango.CmdArgType.DevEncoded):
                attrvalue = magnitude
            else:
                attrvalue = str(magnitude)
        elif fmt in (DataFormat._1D, DataFormat._2D):
            if PyTango.is_int_type(tgtype):
                # cast to integer because the magnitude conversion gives floats
                attrvalue = magnitude.astype('int64')
            elif tgtype == PyTango.CmdArgType.DevUChar:
                attrvalue = magnitude.view('uint8')
            else:
                attrvalue = magnitude
        else:
            attrvalue = str(magnitude)
        return attrvalue

    def decode(self, attr_value):
        """Decodes a value that was received from PyTango into the expected
        representation"""
        if self._pytango_attrinfoex is None:
            self.getAttributeInfoEx(cache=False)
            self._decodeAttrInfoEx()
        value = TangoAttrValue(pytango_dev_attr=attr_value, attr=self)
        return value

    def write(self, value, with_read=True):
        """ Write the value in the Tango Device Attribute """
        try:
            dev = self.getParentObj()
            name, value = self.getSimpleName(), self.encode(value)
            if self.isUsingEvents() or not self.isReadWrite():
                with_read = False
            if with_read:
                try:
                    result = dev.write_read_attribute(name, value)
                except AttributeError:
                    # handle old PyTango
                    dev.write_attribute(name, value)
                    result = dev.read_attribute(name)
                except PyTango.DevFailed as df:
                    for err in df.args:
                        # Handle old device servers
                        if err.reason == 'API_UnsupportedFeature':
                            dev.write_attribute(name, value)
                            result = dev.read_attribute(name)
                            break
                    else:
                        raise df
                self.poll(single=False, value=result, time=time.time())
                return self.decode(result)
            else:
                dev.write_attribute(name, value)
                return None
        except PyTango.DevFailed as df:
            err = df.args[0]
            self.error("[Tango] write failed (%s): %s" %
                       (err.reason, err.desc))
            raise df
        except Exception as e:
            self.error("[Tango] write failed: %s" % str(e))
            raise e

    def poll(self,  single=True, value=None, time=None, error=None):
        """ Notify listeners when the attribute has been polled"""
        with self.__read_lock:
            try:
                if single:
                    self.read(cache=False)
                else:
                    value = self.decode(value)
                    self.__attr_err = error
                    if self.__attr_err:
                        raise self.__attr_err
                    # Avoid "valid-but-outdated" notifications
                    # if FILTER_OLD_TANGO_EVENTS is enabled
                    # and the given timestamp is older than the timestamp
                    # of the cached value
                    filter_old_event = getattr(tauruscustomsettings,
                                               'FILTER_OLD_TANGO_EVENTS',
                                               False)
                    if (self.__attr_value is not None
                            and filter_old_event
                            and time is not None
                            and time < self.__attr_value.time.totime()
                       ):
                        return
                    self.__attr_value = value
            except PyTango.DevFailed as df:
                self.__subscription_event.set()
                self.debug("Error polling: %s" % df.args[0].desc)
                self.traceback()
                self.fireEvent(TaurusEventType.Error, self.__attr_err)
            except Exception as e:
                self.__subscription_event.set()
                self.debug("Error polling: %s" % str(e))
                self.fireEvent(TaurusEventType.Error, self.__attr_err)
            else:
                self.__subscription_event.set()
                self.fireEvent(TaurusEventType.Periodic, self.__attr_value)

    def read(self, cache=True):
        """ Returns the current value of the attribute.
            if cache is set to True (default) or the attribute has events
            active then it will return the local cached value. Otherwise it will
            read the attribute value from the tango device."""
        curr_time = time.time()
        if cache:
            try:
                attr_timestamp = self.__attr_value.time.totime()
            except AttributeError:
                attr_timestamp = 0
            dt = (curr_time - attr_timestamp) * 1000
            if dt < self.getPollingPeriod():
                if self.__attr_value is not None:
                    return self.__attr_value
                elif self.__attr_err is not None:
                    self.debug("[Tango] read from cache failed (%s): %s",
                               self.fullname, self.__attr_err)
                    raise self.__attr_err

        if not cache or (self.__subscription_state in 
                         (SubscriptionState.PendingSubscribe,
                          SubscriptionState.Unsubscribed)
                         and not self.isPollingActive()):
            with self.__read_lock:
                try:
                    dev = self.getParentObj()
                    v = dev.read_attribute(self.getSimpleName())
                    self.__attr_value = self.decode(v)
                    self.__attr_err = None
                    return self.__attr_value
                except PyTango.DevFailed as df:
                    self.__attr_value = None
                    self.__attr_err = df
                    err = df.args[0]
                    self.debug("[Tango] read failed (%s): %s",
                               err.reason, err.desc)
                    raise df
                except Exception as e:
                    self.__attr_value = None
                    self.__attr_err = e
                    self.debug("[Tango] read failed: %s", e)
                    raise e
        elif self.__subscription_state in (SubscriptionState.Subscribing,
                                           SubscriptionState.PendingSubscribe):
            self.__subscription_event.wait()

        if self.__attr_err is not None:
            raise self.__attr_err
        return self.__attr_value


    def getAttributeProxy(self):
        """Convenience method that creates and returns a PyTango.AttributeProxy
        object"""
        return PyTango.AttributeProxy(self.getFullName())

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for listeners
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def __fireRegisterEvent(self, listener):
        '''fire the first config and change (or error) events'''
        try:
            v = self.read()
            # note: it may seem redundant, but some widgets may only react to
            # one or another type, so we should send both for bck-compat
            # Taurus4 widgets should never use config events since the same info
            # is always emitted in a change event
            self.fireEvent(TaurusEventType.Config, v, listener)
            self.fireEvent(TaurusEventType.Change, v, listener)
        except:
            self.fireEvent(TaurusEventType.Error, self.__attr_err, listener)

    def addListener(self, listener):
        """ Add a TaurusListener object in the listeners list.
            If it is the first element and Polling is enabled starts the
            polling mechanism.
            If the listener is already registered nothing happens."""

        listeners = self._listeners
        initial_subscription_state = self.__subscription_state

        ret = TaurusAttribute.addListener(self, listener)
        if not ret:
            return ret

        assert len(listeners) >= 1

        if self.__subscription_state == SubscriptionState.Unsubscribed and len(listeners) == 1:
            self._subscribeEvents()

        # if initial_subscription_state == SubscriptionState.Subscribed:
        if (len(listeners) > 1
            and (initial_subscription_state == SubscriptionState.Subscribed 
                 or self.isPollingActive())
           ):
            sm = self._serialization_mode
            if sm == TaurusSerializationMode.TangoSerial:
                self.__fireRegisterEvent((listener,))
            else:
                job = _BoundMethodWeakrefWithCall(self.__fireRegisterEvent)
                Manager().enqueueJob(job,
                                     job_args=((listener,),),
                                     serialization_mode=sm)
        return ret

    def removeListener(self, listener):
        """ Remove a TaurusListener from the listeners list. If polling enabled
            and it is the last element the stop the polling timer.
            If the listener is not registered nothing happens."""
        ret = TaurusAttribute.removeListener(self, listener)

        if not ret:
            return ret

        if self.hasListeners():
            return ret

        if self.__subscription_state != SubscriptionState.Unsubscribed:
            self._unsubscribeEvents()

        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for attribute configuration
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setConfigEx(self, config):
        self.getParentObj().set_attribute_config([config])

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # PyTango event handling (private)
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def isUsingEvents(self):
        return self.__subscription_state == SubscriptionState.Subscribed
    
    def getSubscriptionState(self):
        return self.__subscription_state    

    def _process_event_exception(self, ex):
        pass

    def _subscribeEvents(self):
        """ Enable subscription to the attribute events. If change events are
            not supported polling is activated """
            
        if self.__chg_evt_id is not None:
            self.warning("chg events already subscribed (id=%s)"
                       %self.__chg_evt_id)
            return

        if self.__dev_hw_obj is None:
            dev = self.getParentObj()
            if dev is None:
                self.debug("failed to subscribe to chg events: device is None")
                return
            self.__dev_hw_obj = dev.getDeviceProxy()
            if self.__dev_hw_obj is None:
                self.debug("failed to subscribe to chg events: HW is None")
                return

        if not self.factory().is_tango_subscribe_enabled():
            self.enablePolling(True)
            return       

        try:
            self.__subscription_state = SubscriptionState.Subscribing
            self._call_dev_hw_subscribe_event(False)
        except:
            self.__subscription_state = SubscriptionState.PendingSubscribe
            self._activatePolling()
            self._call_dev_hw_subscribe_event(True)
                
    def _call_dev_hw_subscribe_event(self, stateless=True):
        """ Executes event subscription on parent TangoDevice objectName
        """
        
        if self.__chg_evt_id is not None:
            self.warning("chg events already subscribed (id=%s)",
                         self.__chg_evt_id)
            return
                
        attr_name = self.getSimpleName()

        # connects to self.push_event callback
        # TODO: _BoundMethodWeakrefWithCall is used as workaround for
        # PyTango #185 issue
        self.__chg_evt_id = self.__dev_hw_obj.subscribe_event(
                attr_name, PyTango.EventType.CHANGE_EVENT,
                _BoundMethodWeakrefWithCall(self.push_event), [], stateless)
        
        return self.__chg_evt_id
                
    def _unsubscribeEvents(self):
        # Careful in this method: This is intended to be executed in the cleanUp
        # so we should not access external objects from the factory, like the
        # parent object
        
        if self.__dev_hw_obj is not None and self.__chg_evt_id is not None:
            self.trace("Unsubscribing to change events (ID=%d)",
                       self.__chg_evt_id)
            try:
                self.__dev_hw_obj.unsubscribe_event(self.__chg_evt_id)
                self.__chg_evt_id = None
            except PyTango.DevFailed as df:
                if len(df.args) and df.args[0].reason == 'API_EventNotFound':
                    # probably tango shutdown has been initiated before and
                    # it unsubscribed from events itself
                    pass
                else:
                    self.debug("Failed: %s", df.args[0].desc)
                    self.trace(str(df))
        self._deactivatePolling()
        self.__subscription_state = SubscriptionState.Unsubscribed

    def _subscribeConfEvents(self):
        """ Enable subscription to the attribute configuration events."""
        self.trace("Subscribing to configuration events...")

        if self.__cfg_evt_id is not None:
            self.warning("cfg events already subscribed (id=%s)"
                       %self.__cfg_evt_id)
            return
        
        if self.__dev_hw_obj is None:
            dev = self.getParentObj()
            if dev is None:
                self.debug("failed to subscribe to cfg events: device is None")
                return
            self.__dev_hw_obj = dev.getDeviceProxy()
            if self.__dev_hw_obj is None:
                self.debug("failed to subscribe to cfg events: HW is None")
                return

        attr_name = self.getSimpleName()
        try:
            # connects to self.push_event callback
            # TODO: _BoundMethodWeakrefWithCall is used as workaround for
            # PyTango #185 issue
            self.__cfg_evt_id = self.__dev_hw_obj.subscribe_event(
                attr_name,
                PyTango.EventType.ATTR_CONF_EVENT,
                _BoundMethodWeakrefWithCall(self.push_event), [], True)
        except PyTango.DevFailed as e:
            self.debug("Error trying to subscribe to CONFIGURATION events.")
            self.traceback()
            # Subscription failed either because event mechanism is not available
            # or because the device server is not running.
            # The first possibility is assumed so an attempt to get the configuration
            # manually is done
            # TODO decide what should be done here
            try:
                attrinfoex = self.__dev_hw_obj.attribute_query(attr_name)
                self._decodeAttrInfoEx(attrinfoex)
            except:
                self.debug("Error getting attribute configuration")
                self.traceback()
                
    def _unsubscribeConfEvents(self):
        # Careful in this method: This is intended to be executed in the cleanUp
        # so we should not access external objects from the factory, like the
        # parent object
        
        if self.__cfg_evt_id is not None and self.__dev_hw_obj is not None:
            self.trace("Unsubscribing to configuration events (ID=%s)",
                       str(self.__cfg_evt_id))
            try:
                self.__dev_hw_obj.unsubscribe_event(self.__cfg_evt_id)
                self.__cfg_evt_id = None
            except PyTango.DevFailed as e:
                self.debug("Error trying to unsubscribe configuration events")
                self.trace(str(e))
                
    def subscribePendingEvents(self):
        """ Execute delayed event subscription
        """                
        if (self.__subscription_state == SubscriptionState.Unsubscribed 
                          or self.isPollingActive()):
            self.__subscription_state = SubscriptionState.PendingSubscribe
        self._subscribeConfEvents()
        self._call_dev_hw_subscribe_event(True)

    def push_event(self, event):
        """Method invoked by the PyTango layer when an event occurs.
        It propagates the event to listeners and delegates other tasks to
        specific handlers for different event types.
        """
        with self.__read_lock:
            # if it is a configuration event
            if isinstance(event, PyTango.AttrConfEventData):
                etype, evalue = self._pushConfEvent(event)
            # if it is an attribute event
            else:
                etype, evalue = self._pushAttrEvent(event)

            # notify the listeners if required (i.e, if etype is not None)
            if etype is None:
                return
            manager = Manager()
            listeners = tuple(self._listeners)
            sm = self._serialization_mode
            if sm == TaurusSerializationMode.TangoSerial:
                self.fireEvent(etype, evalue, listeners=listeners)
            else:
                job = _BoundMethodWeakrefWithCall(self.fireEvent)
                manager.enqueueJob(job, job_args=(etype, evalue),
                                   job_kwargs={'listeners': listeners},
                                   serialization_mode=sm)

    def _pushAttrEvent(self, event):
        """Handler of (non-configuration) events from the PyTango layer.
        It handles the subscription and the (de)activation of polling

        :param event: (A PyTango event)

        :return: (evt_type, evt_value)  Tuple containing the event type and the
                 event value. evt_type is a `TaurusEventType` (or None to
                 indicate that there should not be notification to listeners).
                 evt_value is a TaurusValue, an Exception, or None.
        """
        if not event.err:
            attr_value = self.decode(event.attr_value)
            filter_old_event = getattr(tauruscustomsettings,
                                       'FILTER_OLD_TANGO_EVENTS', False)
            time = event.attr_value.time.totime()

            # Discard "valid" events if the attribute value is not None
            # and FILTER_OLD_TANGO_EVENTS is enabled
            # and the given timestamp is older than the timestamp
            # of the cache value
            if (self.__attr_value is not None
                and self.__subscription_state == SubscriptionState.Subscribed
                and filter_old_event
                and time < self.__attr_value.time.totime()):
                return [None, None]

            self.__attr_value = attr_value
            self.__attr_err = None
            self.__subscription_state = SubscriptionState.Subscribed
            self.__subscription_event.set()
            if not self.isPollingForced():
                self._deactivatePolling()
            return TaurusEventType.Change, self.__attr_value

        elif event.errors[0].reason in EVENT_TO_POLLING_EXCEPTIONS:
            if not self.isPollingActive():
                self.info("Activating polling. Reason: %s",
                          event.errors[0].reason)
                self.__subscription_state = SubscriptionState.PendingSubscribe
                self._activatePolling()
            return [None, None]

        else:
            self.__attr_value, self.__attr_err = None, PyTango.DevFailed(
                *event.errors)
            self.__subscription_state = SubscriptionState.Subscribed
            self.__subscription_event.set()
            self._deactivatePolling()
            return TaurusEventType.Error, self.__attr_err

    def _pushConfEvent(self, event):
        """Handler of AttrConfEventData events from the PyTango layer.

        :param event: (PyTango.AttrConfEventData)

        :return: (evt_type, evt_value)  Tuple containing the event type and the
                 event value. evt_type is a `TaurusEventType` (or None to
                 indicate that there should not be notification to listeners).
                 evt_value is a TaurusValue, an Exception, or None.
        """
        if not event.err:
            # update conf-related attributes
            self._decodeAttrInfoEx(event.attr_conf)
            # make sure that there is a self.__attr_value
            if self.__attr_value is None:
                # TODO: maybe we can avoid this read?
                self.__attr_value = self.getValueObj(cache=False)
            return TaurusEventType.Config, self.__attr_value

        else:
            self.__attr_value, self.__attr_err = None, PyTango.DevFailed(
                *event.errors)
            return TaurusEventType.Error, self.__attr_err

    def isWrite(self, cache=True):
        return self.getTangoWritable(cache) == PyTango.AttrWriteType.WRITE

    def isReadOnly(self, cache=True):
        return not self.getTangoWritable(cache) == PyTango.AttrWriteType.READ

    def isReadWrite(self, cache=True):
        return self.getTangoWritable(cache) == PyTango.AttrWriteType.READ_WRITE

    def getTangoWritable(self, cache=True):
        """like TaurusAttribute.isWritable(), but it returns a
        PyTango.AttrWriteType instead of a bool"""
        return self.tango_writable

    def getLabel(self, cache=True):
        return self._label

    def getRange(self, cache=True):
        return self._range

    def getLimits(self, cache=True):
        return self.getRange(cache)

    def getRanges(self, cache=True):
        return self.getRange()

    def getAlarms(self, cache=True):
        return self._alarm

    def getWarnings(self, cache=True):
        return self._warning

    def getMaxDim(self, cache=True):
        return self.max_dim_x, self.max_dim_y

    def setLimits(self, low, high):
        self.setRange([low, high])

    def setLabel(self, lbl):
        TaurusAttribute.setLabel(self, lbl)
        infoex = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        infoex.label = lbl
        self._applyConfig()

    def setRange(self, *limits):
        if isinstance(limits[0], list):
            limits = limits[0]
        low, high = limits
        low = Quantity(low)
        if low.unitless:
            low = Quantity(low.magnitude, self._units)
        high = Quantity(high)
        if high.unitless:
            high = Quantity(high.magnitude, self._units)
        TaurusAttribute.setRange(self, [low, high])
        infoex = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        if low.magnitude != float('-inf'):
            infoex.min_value = str(low.to(self._units).magnitude)
        else:
            infoex.min_value = 'Not specified'
        if high.magnitude != float('inf'):
            infoex.max_value = str(high.to(self._units).magnitude)
        else:
            infoex.max_value = 'Not specified'
        self._applyConfig()

    def setWarnings(self, *limits):
        if isinstance(limits[0], list):
            limits = limits[0]
        low, high = limits
        low = Quantity(low)
        if low.unitless:
            low = Quantity(low.magnitude, self._units)
        high = Quantity(high)
        if high.unitless:
            high = Quantity(high.magnitude, self._units)
        TaurusAttribute.setWarnings(self, [low, high])
        infoex = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        if low.magnitude != float('-inf'):
            infoex.alarms.min_warning = str(low.to(self._units).magnitude)
        else:
            infoex.alarms.min_warning = 'Not specified'
        if high.magnitude != float('inf'):
            infoex.alarms.max_warning = str(high.to(self._units).magnitude)
        else:
            infoex.alarms.max_warning = 'Not specified'
        self._applyConfig()

    def setAlarms(self, *limits):
        if isinstance(limits[0], list):
            limits = limits[0]
        low, high = limits
        low = Quantity(low)
        if low.unitless:
            low = Quantity(low.magnitude, self._units)
        high = Quantity(high)
        if high.unitless:
            high = Quantity(high.magnitude, self._units)
        TaurusAttribute.setAlarms(self, [low, high])
        infoex = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        if low.magnitude != float('-inf'):
            infoex.alarms.min_alarm = str(low.to(self._units).magnitude)
        else:
            infoex.alarms.min_alarm = 'Not specified'
        if high.magnitude != float('inf'):
            infoex.alarms.max_alarm = str(high.to(self._units).magnitude)
        else:
            infoex.alarms.max_alarm = 'Not specified'
        self._applyConfig()

    def _applyConfig(self):
        config = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        self.setConfigEx(config)

    def _decodeAttrInfoEx(self, pytango_attrinfoex=None):
        if pytango_attrinfoex is None:
            return

        self._pytango_attrinfoex = i = pytango_attrinfoex

        self.writable = i.writable != PyTango.AttrWriteType.READ
        self._label = i.label
        self.data_format = data_format_from_tango(i.data_format)
        desc = description_from_tango(i.description)
        if desc != "":
            self._description = desc
        self.type = data_type_from_tango(i.data_type)
        ###############################################################
        # changed in taurus4: range, alarm and warning now return
        # quantities if appropriate
        if self.isNumeric():
            units = self._unit_from_tango(i.unit)
        else:
            units = UR.parse_units(None)

        if PyTango.is_numerical_type(i.data_type, inc_array=True):
            Q_ = partial(quantity_from_tango_str, units=units,
                         dtype=i.data_type)
            ninf, inf = float('-inf'), float('inf')

            min_value = Q_(i.min_value)
            if min_value is None:
                min_value = Quantity(ninf, units)
            max_value = Q_(i.max_value)
            if max_value is None:
                max_value = Quantity(inf, units)
            min_alarm = Q_(i.alarms.min_alarm)
            if min_alarm is None:
                min_alarm = Quantity(ninf, units)
            max_alarm = Q_(i.alarms.max_alarm)
            if max_alarm is None:
                max_alarm = Quantity(inf, units)
            min_warning = Q_(i.alarms.min_warning)
            if min_warning is None:
                min_warning = Quantity(ninf, units)
            max_warning = Q_(i.alarms.max_warning)
            if max_warning is None:
                max_warning = Quantity(inf, units)

            self._range = [min_value, max_value]
            self._warning = [min_warning, max_warning]
            self._alarm = [min_alarm, max_alarm]

        ###############################################################
        # The following members will be accessed via __getattr__
        # self.standard_unit
        # self.display_unit
        # self.disp_level
        ###############################################################
        # Tango-specific extension of TaurusConfigValue
        self.display_level = display_level_from_tango(i.disp_level)
        self.tango_writable = i.writable
        self.max_dim = i.max_dim_x, i.max_dim_y
        ###############################################################
        fmt = standard_display_format_from_tango(i.data_type, i.format)
        self.format_spec = fmt.lstrip('%')  # format specifier
        match = re.search("[^\.]*\.(?P<precision>[0-9]+)[eEfFgG%]", fmt)
        if match:
            self.precision = int(match.group(1))
        # self._units and self._display_format is to be used by
        # TangoAttrValue for performance reasons. Do not rely on it in other
        # code
        self._units = units

    @property
    @deprecation_decorator(alt='format_spec or precision', rel='4.0.4')
    def format(self):
        i = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        return standard_display_format_from_tango(i.data_type, i.format)

    @property
    def _tango_data_type(self):
        '''returns the *tango* (not Taurus) data type'''
        i = self._pytango_attrinfoex or PyTango.AttributeInfoEx()
        return i.data_type

    def _unit_from_tango(self, unit):
        # silently treat unit-not-defined as unitless
        # TODO: consider logging that unit-not-defined is treated as unitless
        # TODO: See https://github.com/taurus-org/taurus/issues/584 and
        # https://github.com/taurus-org/taurus/pull/662
        # The extra comparison to "No unit" is necessary where
        # server/database runs Tango 7 or 8 and client runs higher versions.
        if unit == PyTango.constants.UnitNotSpec or unit == "No unit":
            unit = None
        try:
            return UR.parse_units(unit)
        except Exception as e:
            # TODO: Maybe we could dynamically register the unit in the UR
            msg = 'Unknown unit "%s" (will be treated as unitless)'
            if self.__already_warned_unit == unit:
                self.debug(msg, unit)
            else:
                self.warning(msg, unit)
                self.debug('%r', e)
                self.__already_warned_unit = unit
            return UR.parse_units(None)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Deprecated methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @taurus4_deprecation(alt=".description")
    def getDescription(self, cache=True):
        return self.description

    @taurus4_deprecation(alt=".description")
    def setDescription(self, descr):
        self.description = descr

    @taurus4_deprecation()
    def isInformDeviceOfErrors(self):
        return False

    @taurus4_deprecation(dbg_msg='Deprecated method')
    def displayValue(self, value):
        return str(value)

    @taurus4_deprecation(alt='getLabel')
    def getDisplayValue(self, cache=True):
        attrvalue = self.getValueObj(cache=cache)
        if not attrvalue:
            return None
        v = attrvalue.rvalue

        return self.displayValue(v)

    @taurus4_deprecation(alt='.rvalue.units')
    def getStandardUnit(self, cache=True):
        try:
            return str(self.getValueObj(cache).rvalue.units)
        except:
            return None

    @taurus4_deprecation(alt='.rvalue.units')
    def getDisplayUnit(self, cache=True):
        try:
            return str(self.getValueObj(cache).rvalue.units)
        except:
            return None

    @taurus4_deprecation(dbg_msg='Do not use')
    def getDisplayWriteValue(self, cache=True):
        raise NotImplementedError("Not available since Taurus4")

    @taurus4_deprecation(alt='isWritable')
    def getWritable(self, cache=True):
        return self.isWritable(cache)

    @taurus4_deprecation(alt='self.data_format')
    def isScalar(self):
        return self.data_format == DataFormat._0D

    @taurus4_deprecation(alt='self.data_format')
    def isSpectrum(self):
        return self.data_format == DataFormat._1D

    @taurus4_deprecation(alt='self.data_format')
    def isImage(self):
        return self.data_format == DataFormat._2D

    @taurus4_deprecation(alt='getMaxDim')
    def getMaxDimX(self, cache=True):
        dim = self.getMaxDim(cache)
        if dim:
            return dim[0]
        else:
            return None

    @taurus4_deprecation(alt='getMaxDim')
    def getMaxDimY(self, cache=True):
        dim = self.getMaxDim(cache)
        if dim:
            return dim[1]
        else:
            return None

    @taurus4_deprecation(dbg_msg='Deprecated method')
    def getShape(self, cache=True):
        if self.isScalar(cache):
            return ()
        elif self.isSpectrum():
            return (self.getMaxDimX(),)
        else:
            return self.getMaxDim()

    @taurus4_deprecation(alt='getAttributeInfoEx')
    def getParam(self, param_name):
        """ Get attributes of AttributeInfoEx (PyTango)
        """
        try:
            return getattr(self._pytango_attrinfoex, param_name)
        except:
            return None

    @taurus4_deprecation(alt='PyTango')
    def setParam(self, param_name, value):
        """ Set attributes of AttributeInfoEx (PyTango)
        """
        if hasattr(self._pytango_attrinfoex, param_name):
            setattr(self._pytango_attrinfoex, param_name, str(value))
        self._applyConfig()

    @taurus4_deprecation(alt='self')
    def getConfig(self):
        """Returns the current configuration of the attribute."""
        return weakref.proxy(self)

    def getAttributeInfoEx(self, cache=True):
        if not cache:
            try:
                attr_name = self.getSimpleName()
                attrinfoex = self.__dev_hw_obj.attribute_query(attr_name)
                self._decodeAttrInfoEx(attrinfoex)
            except Exception as e:
                self.debug("Error getting attribute configuration: %s", e)
                self.traceback()

        return self._pytango_attrinfoex

    @taurus4_deprecation(alt='.rvalue.units')
    def getUnit(self, cache=True):
        try:
            return str(self.getValueObj(cache).rvalue.units)
        except:
            return None

    @taurus4_deprecation(alt='.rvalue.units')
    def _set_unit(self, value):
        '''for backwards compat with taurus < 4'''
        extra_msg = 'Ignoring setting of units of %s to %r' % (self.name,
                                                               value)
        self.debug(extra_msg)

    @taurus4_deprecation(alt='getMinRange')
    def getMinValue(self, cache=True):
        return self.getMinRange()

    @taurus4_deprecation(alt='getMaxRange')
    def getMaxValue(self, cache=True):
        return self.getMaxRange()

    @taurus4_deprecation(alt='getRange')
    def getCLimits(self):
        if self._pytango_attrinfoex is not None:
            value = [self._pytango_attrinfoex.min_value,
                     self._pytango_attrinfoex.max_value]
        else:
            value = [self.not_specified, self.not_specified]
        return value

    @taurus4_deprecation(alt='getAlarms')
    def getCAlarms(self):
        if self._pytango_attrinfoex is not None:
            value = [self._pytango_attrinfoex.min_alarm,
                     self._pytango_attrinfoex.max_alarm]
        else:
            value = [self.not_specified, self.not_specified]
        return value

    @taurus4_deprecation(alt='getWarnings')
    def getCWarnings(self):
        if self._pytango_attrinfoex is not None:
            value = [self._pytango_attrinfoex.alarms.min_warning,
                     self._pytango_attrinfoex.alarms.max_warning]
        else:
            value = [self.not_specified, self.not_specified]
        return value

    @taurus4_deprecation(alt='getRange + getAlarms + getWarnings')
    def getCRanges(self):
        if self._pytango_attrinfoex is not None:
            value = [self._pytango_attrinfoex.min_value,
                     self._pytango_attrinfoex.min_alarm,
                     self._pytango_attrinfoex.alarms.min_warning,
                     self._pytango_attrinfoex.alarms.max_warning,
                     self._pytango_attrinfoex.max_alarm,
                     self._pytango_attrinfoex.max_value]
        else:
            value = [self.not_specified, self.not_specified,
                     self.not_specified, self.not_specified,
                     self.not_specified, self.not_specified]
        return value

    @taurus4_deprecation(alt='.alarms[0]')
    def getMinAlarm(self):
        if self._pytango_attrinfoex is None:
            return None
        return self._pytango_attrinfoex.alarms.min_alarm

    @taurus4_deprecation(alt='.alarms[1]')
    def getMaxAlarm(self):
        if self._pytango_attrinfoex is None:
            return None
        return self._pytango_attrinfoex.alarms.max_alarm

    @taurus4_deprecation(alt='.warnings[0]')
    def getMinWarning(self):
        if self._pytango_attrinfoex is None:
            return None
        return self._pytango_attrinfoex.alarms.min_warning

    @taurus4_deprecation(alt='.warnings[1]')
    def getMaxWarning(self):
        if self._pytango_attrinfoex is None:
            return None
        return self._pytango_attrinfoex.alarms.max_warning

    @taurus4_deprecation(alt='.alarms')
    def setMinAlarm(self, value):
        if self._pytango_attrinfoex is None:
            self._pytango_attrinfoex.alarms.min_alarm = str(value)
            self._applyConfig()

    @taurus4_deprecation(alt='.alarms')
    def setMaxAlarm(self, value):
        if self._pytango_attrinfoex is None:
            self._pytango_attrinfoex.alarms.max_alarm = str(value)
            self._applyConfig()

    @taurus4_deprecation(alt='.warnings')
    def setMinWarning(self, value):
        if self._pytango_attrinfoex is None:
            self._pytango_attrinfoex.alarms.min_warning = str(value)
            self._applyConfig()

    @taurus4_deprecation(alt='.warnings')
    def setMaxWarning(self, value):
        if self._pytango_attrinfoex is None:
            self._pytango_attrinfoex.alarms.max_warning = str(value)
            self._applyConfig()

    # deprecated property!
    unit = property(getUnit, _set_unit)
    climits = property(getCLimits)
    calarms = property(getCAlarms)
    cwarnings = property(getCAlarms)
    cranges = property(getCRanges)

    min_alarm = property(getMinAlarm, setMinAlarm)
    max_alarm = property(getMaxAlarm, setMaxAlarm)
    min_warning = property(getMinWarning, setMinWarning)
    max_warning = property(getMaxWarning, setMaxWarning)

    # properties
    label = property(getLabel, setLabel)
    range = property(getRange, setRange)
    warnings = property(getWarnings, setWarnings)
    alarms = property(getAlarms, setAlarms)

    @property
    def description(self):
        return self._description

    @property
    def dev_alias(self):
        self.deprecated(dep='dev_alias', alt='getParentObj().name', rel='tep14')
        parent = self.getParentObj()
        if parent is None:
            return None
        else:
            return parent.name

    @description.setter
    def description(self, descr):
        if descr != self._description:
            if descr == '':
                descr = 'A Tango Attribute'
            self._description = descr
            self._applyConfig()


class TangoAttributeEventListener(EventListener):
    """A class that listens for an event with a specific value

    Note: Since this class stores for each event value the last timestamp when
    it occured, it should only be used for events for which the event value
    domain (possible values) is limited and well known (ex: an enum)"""

    def __init__(self, attr):
        EventListener.__init__(self)
        self.attr = attr
        attr.addListener(self)

    def eventReceived(self, s, t, v):
        if t not in (TaurusEventType.Change, TaurusEventType.Periodic):
            return
        self.fireEvent(v.value)


def test1():
    import numpy
    from taurus import Attribute
    a = Attribute('sys/tg_test/1/ulong64_scalar')

    a.write(numpy.uint64(88))

if __name__ == "__main__":
    test1()
